#!/usr/bin/env python

#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Module implementing the command line entry point for executing tests.

This module can be executed from the command line using the following
approaches::

    python -m robot.run
    python path/to/robot/run.py

Instead of ``python`` it is possible to use also other Python interpreters.
This module is also used by the installed ``robot`` start-up script.

This module also provides :func:`run` and :func:`run_cli` functions
that can be used programmatically. Other code is for internal usage.
"""

import sys

if __name__ == '__main__' and 'robot' not in sys.modules:
    import pythonpathsetter

from robot.conf import RobotSettings
from robot.model import ModelModifier
from robot.output import LOGGER, pyloggingconf
from robot.reporting import ResultWriter
from robot.running.builder import TestSuiteBuilder
from robot.utils import Application, text


USAGE = """Robot Framework -- A generic automation framework

Version:  <VERSION>

Usage:  robot [options] paths
   or:  python -m robot [options] paths
   or:  python path/to/robot [options] paths

Robot Framework is a generic open source automation framework for acceptance
testing, acceptance test-driven development (ATDD) and robotic process
automation (RPA). It has simple, easy-to-use syntax that utilizes the
keyword-driven automation approach. Keywords adding new capabilities are
implemented in libraries using Python. New higher level
keywords can also be created using Robot Framework's own syntax.

The easiest way to execute Robot Framework is using the `robot` command created
as part of the normal installation. Alternatively it is possible to execute
the `robot` module directly like `python -m robot`, where `python` can be
replaced with any supported Python interpreter. Yet another alternative
is running the `robot` directory like `python path/to/robot`.

Tests (or tasks in RPA terminology) are created in files typically having the
`*.robot` extension. Files automatically create test (or task) suites and
directories with these files create higher level suites. When Robot Framework
is executed, paths to these files or directories are given to it as arguments.

By default Robot Framework creates an XML output file and a log and a report in
HTML format, but this can be configured using various options listed below.
Outputs in HTML format are for human consumption and XML output for integration
with other systems. XML outputs can also be combined and otherwise further
post-processed with the Rebot tool that is an integral part of Robot Framework.
Run `rebot --help` for more information.

Robot Framework is open source software released under Apache License 2.0.
For more information about the framework and the rich ecosystem around it
see http://robotframework.org/.

Options
=======

    --rpa                 Turn on the generic automation mode. Mainly affects
                          terminology so that "test" is replaced with "task"
                          in logs and reports. By default the mode is got
                          from test/task header in data files.
    --language lang *     Activate localization. `lang` can be a name or a code
                          of a built-in language, or a path or a module name of
                          a custom language file.
 -F --extension value     Parse only files with this extension when executing
                          a directory. Has no effect when running individual
                          files or when using resource files. If more than one
                          extension is needed, separate them with a colon.
                          Examples: `--extension txt`, `--extension robot:txt`
                          Only `*.robot` files are parsed by default.
 -I --parseinclude pattern *  Parse only files matching `pattern`. It can be:
                          - a file name or pattern like `example.robot` or
                            `*.robot` to parse all files matching that name,
                          - a file path like `path/to/example.robot`, or
                          - a directory path like `path/to/example` to parse
                            all files in that directory, recursively.
 -N --name name           Set the name of the top level suite. By default the
                          name is created based on the executed file or
                          directory.
 -D --doc documentation   Set the documentation of the top level suite.
                          Simple formatting is supported (e.g. *bold*). If the
                          documentation contains spaces, it must be quoted.
                          If the value is path to an existing file, actual
                          documentation is read from that file.
                          Examples: --doc "Very *good* example"
                                    --doc doc_from_file.txt
 -M --metadata name:value *  Set metadata of the top level suite. Value can
                          contain formatting and be read from a file similarly
                          as --doc. Example: --metadata Version:1.2
 -G --settag tag *        Sets given tag(s) to all executed tests.
 -t --test name *         Select tests by name or by long name containing also
                          parent suite name like `Parent.Test`. Name is case
                          and space insensitive and it can also be a simple
                          pattern where `*` matches anything, `?` matches any
                          single character, and `[chars]` matches one character
                          in brackets.
    --task name *         Alias to --test. Especially applicable with --rpa.
 -s --suite name *        Select suites by name. When this option is used with
                          --test, --include or --exclude, only tests in
                          matching suites and also matching other filtering
                          criteria are selected. Name can be a simple pattern
                          similarly as with --test and it can contain parent
                          name separated with a dot. For example, `-s X.Y`
                          selects suite `Y` only if its parent is `X`.
 -i --include tag *       Select tests by tag. Similarly as name with --test,
                          tag is case and space insensitive and it is possible
                          to use patterns with `*`, `?` and `[]` as wildcards.
                          Tags and patterns can also be combined together with
                          `AND`, `OR`, and `NOT` operators.
                          Examples: --include foo --include bar*
                                    --include fooANDbar*
 -e --exclude tag *       Select test cases not to run by tag. These tests are
                          not run even if included with --include. Tags are
                          matched using same rules as with --include.
 -R --rerunfailed output  Select failed tests from an earlier output file to be
                          re-executed. Equivalent to selecting same tests
                          individually using --test.
 -S --rerunfailedsuites output  Select failed suites from an earlier output
                          file to be re-executed.
    --runemptysuite       Executes suite even if it contains no tests. Useful
                          e.g. with --include/--exclude when it is not an error
                          that no test matches the condition.
    --skip tag *          Tests having given tag will be skipped. Tag can be
                          a pattern.
    --skiponfailure tag *  Tests having given tag will be skipped if they fail.
                          Tag can be a pattern
 -v --variable name:value *  Set variables in the test data. Only scalar
                          variables with string value are supported and name is
                          given without `${}`. See --variablefile for a more
                          powerful variable setting mechanism.
                          Examples:
                          --variable name:Robot  =>  ${name} = `Robot`
                          -v "hello:Hello world" =>  ${hello} = `Hello world`
                          -v x: -v y:42          =>  ${x} = ``, ${y} = `42`
 -V --variablefile path *  Python or YAML file file to read variables from.
                          Possible arguments to the variable file can be given
                          after the path using colon or semicolon as separator.
                          Examples: --variablefile path/vars.yaml
                                    --variablefile environment.py:testing
 -d --outputdir dir       Where to create output files. The default is the
                          directory where tests are run from and the given path
                          is considered relative to that unless it is absolute.
 -o --output file         XML output file. Given path, similarly as paths given
                          to --log, --report, --xunit, and --debugfile, is
                          relative to --outputdir unless given as an absolute
                          path. Other output files are created based on XML
                          output files after the test execution and XML outputs
                          can also be further processed with Rebot tool. Can be
                          disabled by giving a special value `NONE`.
                          Default: output.xml
 -l --log file            HTML log file. Can be disabled by giving a special
                          value `NONE`. Default: log.html
                          Examples: `--log mylog.html`, `-l NONE`
 -r --report file         HTML report file. Can be disabled with `NONE`
                          similarly as --log. Default: report.html
 -x --xunit file          xUnit compatible result file. Not created unless this
                          option is specified.
 -b --debugfile file      Debug file written during execution. Not created
                          unless this option is specified.
 -T --timestampoutputs    When this option is used, timestamp in a format
                          `YYYYMMDD-hhmmss` is added to all generated output
                          files between their basename and extension. For
                          example `-T -o output.xml -r report.html -l none`
                          creates files like `output-20070503-154410.xml` and
                          `report-20070503-154410.html`.
    --splitlog            Split the log file into smaller pieces that open in
                          browsers transparently.
    --logtitle title      Title for the generated log file. The default title
                          is `<SuiteName> Log`.
    --reporttitle title   Title for the generated report file. The default
                          title is `<SuiteName> Report`.
    --reportbackground colors  Background colors to use in the report file.
                          Given in format `passed:failed:skipped` where the
                          `:skipped` part can be omitted. Both color names and
                          codes work.
                          Examples: --reportbackground green:red:yellow
                                    --reportbackground #00E:#E00
    --maxerrorlines lines  Maximum number of error message lines to show in
                          report when tests fail. Default is 40, minimum is 10
                          and `NONE` can be used to show the full message.
    --maxassignlength characters  Maximum number of characters to show in log
                          when variables are assigned. Zero or negative values
                          can be used to avoid showing assigned values at all.
                          Default is 200.
 -L --loglevel level      Threshold level for logging. Available levels: TRACE,
                          DEBUG, INFO (default), WARN, NONE (no logging). Use
                          syntax `LOGLEVEL:DEFAULT` to define the default
                          visible log level in log files.
                          Examples: --loglevel DEBUG
                                    --loglevel DEBUG:INFO
    --suitestatlevel level  How many levels to show in `Statistics by Suite`
                          in log and report. By default all suite levels are
                          shown. Example:  --suitestatlevel 3
    --tagstatinclude tag *  Include only matching tags in `Statistics by Tag`
                          in log and report. By default all tags are shown.
                          Given tag can be a pattern like with --include.
    --tagstatexclude tag *  Exclude matching tags from `Statistics by Tag`.
                          This option can be used with --tagstatinclude
                          similarly as --exclude is used with --include.
    --tagstatcombine tags:name *  Create combined statistics based on tags.
                          These statistics are added into `Statistics by Tag`.
                          If the optional `name` is not given, name of the
                          combined tag is got from the specified tags. Tags are
                          matched using the same rules as with --include.
                          Examples: --tagstatcombine requirement-*
                                    --tagstatcombine tag1ANDtag2:My_name
    --tagdoc pattern:doc *  Add documentation to tags matching the given
                          pattern. Documentation is shown in `Test Details` and
                          also as a tooltip in `Statistics by Tag`. Pattern can
                          use `*`, `?` and `[]` as wildcards like --test.
                          Documentation can contain formatting like --doc.
                          Examples: --tagdoc mytag:Example
                                    --tagdoc "owner-*:Original author"
    --tagstatlink pattern:link:title *  Add external links into `Statistics by
                          Tag`. Pattern can use `*`, `?` and `[]` as wildcards
                          like --test. Characters matching to `*` and `?`
                          wildcards can be used in link and title with syntax
                          %N, where N is index of the match (starting from 1).
                          Examples: --tagstatlink mytag:http://my.domain:Title
                          --tagstatlink "bug-*:http://url/id=%1:Issue Tracker"
    --expandkeywords name:<pattern>|tag:<pattern> *
                          Matching keywords will be automatically expanded in
                          the log file. Matching against keyword name or tags
                          work using same rules as with --removekeywords.
                          Examples: --expandkeywords name:BuiltIn.Log
                                    --expandkeywords tag:expand
    --removekeywords all|passed|for|wuks|name:<pattern>|tag:<pattern> *
                          Remove keyword data from the generated log file.
                          Keywords containing warnings are not removed except
                          in the `all` mode.
                          all:     remove data from all keywords
                          passed:  remove data only from keywords in passed
                                   test cases and suites
                          for:     remove passed iterations from for loops
                          while:   remove passed iterations from while loops
                          wuks:    remove all but the last failing keyword
                                   inside `BuiltIn.Wait Until Keyword Succeeds`
                          name:<pattern>:  remove data from keywords that match
                                   the given pattern. The pattern is matched
                                   against the full name of the keyword (e.g.
                                   'MyLib.Keyword', 'resource.Second Keyword'),
                                   is case, space, and underscore insensitive,
                                   and may contain `*`, `?` and `[]` wildcards.
                                   Examples: --removekeywords name:Lib.HugeKw
                                             --removekeywords name:myresource.*
                          tag:<pattern>:  remove data from keywords that match
                                   the given pattern. Tags are case and space
                                   insensitive and patterns can contain `*`,
                                   `?` and `[]` wildcards. Tags and patterns
                                   can also be combined together with `AND`,
                                   `OR`, and `NOT` operators.
                                   Examples: --removekeywords foo
                                             --removekeywords fooANDbar*
    --flattenkeywords for|while|iteration|name:<pattern>|tag:<pattern> *
                          Flattens matching keywords in the generated log file.
                          Matching keywords get all log messages from their
                          child keywords and children are discarded otherwise.
                          for:     flatten FOR loops fully
                          while:   flatten WHILE loops fully
                          iteration: flatten FOR/WHILE loop iterations
                          foritem: deprecated alias for `iteration`
                          name:<pattern>:  flatten matched keywords using same
                                   matching rules as with
                                   `--removekeywords name:<pattern>`
                          tag:<pattern>:  flatten matched keywords using same
                                   matching rules as with
                                   `--removekeywords tag:<pattern>`
    --nostatusrc          Sets the return code to zero regardless of failures
                          in test cases. Error codes are returned normally.
    --dryrun              Verifies test data and runs tests so that library
                          keywords are not executed.
 -X --exitonfailure       Stops test execution if any test fails.
    --exitonerror         Stops test execution if any error occurs when parsing
                          test data, importing libraries, and so on.
    --skipteardownonexit  Causes teardowns to be skipped if test execution is
                          stopped prematurely.
    --randomize all|suites|tests|none  Randomizes the test execution order.
                          all:    randomizes both suites and tests
                          suites: randomizes suites
                          tests:  randomizes tests
                          none:   no randomization (default)
                          Use syntax `VALUE:SEED` to give a custom random seed.
                          The seed must be an integer.
                          Examples: --randomize all
                                    --randomize tests:1234
    --listener listener *  Class or module for monitoring test execution.
                          Gets notifications e.g. when tests start and end.
                          Arguments to the listener class can be given after
                          the name using a colon or a semicolon as a separator.
                          Examples: --listener MyListener
                                    --listener path/to/Listener.py:arg1:arg2
    --prerunmodifier modifier *  Class to programmatically modify the suite
                          structure before execution. Accepts arguments the
                          same way as with --listener.
    --prerebotmodifier modifier *  Class to programmatically modify the result
                          model before creating reports and logs. Accepts
                          arguments the same way as with --listener.
    --parser parser *     Custom parser class or module. Parser classes accept
                          arguments the same way as with --listener.
    --console type        How to report execution on the console.
                          verbose:  report every suite and test (default)
                          dotted:   only show `.` for passed test, `s` for
                                    skipped tests, and `F` for failed tests
                          quiet:    no output except for errors and warnings
                          none:     no output whatsoever
 -. --dotted              Shortcut for `--console dotted`.
    --quiet               Shortcut for `--console quiet`.
 -W --consolewidth chars  Width of the console output. Default is 78.
 -C --consolecolors auto|on|ansi|off  Use colors on console output or not.
                          auto: use colors when output not redirected (default)
                          on:   always use colors
                          ansi: like `on` but use ANSI colors also on Windows
                          off:  disable colors altogether
 -K --consolemarkers auto|on|off  Show markers on the console when top level
                          keywords in a test case end. Values have same
                          semantics as with --consolecolors.
 -P --pythonpath path *   Additional locations (directories, ZIPs) where to
                          search libraries and other extensions when they are
                          imported. Multiple paths can be given by separating
                          them with a colon (`:`) or by using this option
                          several times. Given path can also be a glob pattern
                          matching multiple paths.
                          Examples: --pythonpath libs/
                                    --pythonpath /opt/libs:libraries.zip
 -A --argumentfile path *  Text file to read more arguments from. Use special
                          path `STDIN` to read contents from the standard input
                          stream. File can have both options and input files
                          or directories, one per line. Contents do not need to
                          be escaped but spaces in the beginning and end of
                          lines are removed. Empty lines and lines starting
                          with a hash character (#) are ignored.
                          Example file:
                          |  --include regression
                          |  --name Regression Tests
                          |  # This is a comment line
                          |  tests.robot
                          |  path/to/test/directory/
                          Examples:
                          --argumentfile argfile.txt --argumentfile STDIN
 -h -? --help             Print usage instructions.
 --version                Print version information.

Options that are marked with an asterisk (*) can be specified multiple times.
For example, `--test first --test third` selects test cases with name `first`
and `third`. If an option accepts a value but is not marked with an asterisk,
the last given value has precedence. For example, `--log A.html --log B.html`
creates log file `B.html`. Options accepting no values can be disabled by
using the same option again with `no` prefix added or dropped. The last option
has precedence regardless of how many times options are used. For example,
`--dryrun --dryrun --nodryrun --nostatusrc --statusrc` would not activate the
dry-run mode and would return a normal return code.

Long option names are case and hyphen insensitive. For example, --TagStatLink
and --tag-stat-link are equivalent to, but easier to read than, --tagstatlink.
Long options can also be shortened as long as they are unique.

Environment Variables
=====================

ROBOT_OPTIONS             Space separated list of default options to be placed
                          in front of any explicit options on the command line.
ROBOT_SYSLOG_FILE         Path to a file where Robot Framework writes internal
                          information about parsing test case files and running
                          tests. Can be useful when debugging problems. If not
                          set, or set to a special value `NONE`, writing to the
                          syslog file is disabled.
ROBOT_SYSLOG_LEVEL        Log level to use when writing to the syslog file.
                          Available levels are the same as with --loglevel
                          command line option and the default is INFO.
ROBOT_INTERNAL_TRACES     When set to any non-empty value, Robot Framework's
                          internal methods are included in error tracebacks.

Examples
========

# Simple test run using `robot` command without options.
$ robot tests.robot

# Using options.
$ robot --include smoke --name "Smoke Tests" path/to/tests.robot

# Executing `robot` module using Python.
$ python -m robot path/to/tests

# Executing multiple test case files and using case-insensitive long options.
$ robot --SuiteStatLevel 2 --Metadata Version:3 tests/*.robot more/tests.robot

# Setting default options and syslog file before running tests.
$ export ROBOT_OPTIONS="--outputdir results --suitestatlevel 2"
$ export ROBOT_SYSLOG_FILE=/tmp/syslog.txt
$ robot tests.robot
"""


class RobotFramework(Application):

    def __init__(self):
        super().__init__(USAGE, arg_limits=(1,), env_options='ROBOT_OPTIONS',
                         logger=LOGGER)

    def main(self, datasources, **options):
        try:
            settings = RobotSettings(options)
        except:
            LOGGER.register_console_logger(stdout=options.get('stdout'),
                                           stderr=options.get('stderr'))
            raise
        LOGGER.register_console_logger(**settings.console_output_config)
        LOGGER.info(f'Settings:\n{settings}')
        if settings.pythonpath:
            sys.path = settings.pythonpath + sys.path
        builder = TestSuiteBuilder(included_extensions=settings.extension,
                                   included_files=settings.parse_include,
                                   custom_parsers=settings.parsers,
                                   rpa=settings.rpa,
                                   lang=settings.languages,
                                   allow_empty_suite=settings.run_empty_suite)
        suite = builder.build(*datasources)
        if settings.pre_run_modifiers:
            suite.visit(ModelModifier(settings.pre_run_modifiers,
                                      settings.run_empty_suite, LOGGER))
        suite.configure(**settings.suite_config)
        settings.rpa = suite.validate_execution_mode()
        with pyloggingconf.robot_handler_enabled(settings.log_level):
            old_max_error_lines = text.MAX_ERROR_LINES
            old_max_assign_length = text.MAX_ASSIGN_LENGTH
            text.MAX_ERROR_LINES = settings.max_error_lines
            text.MAX_ASSIGN_LENGTH = settings.max_assign_length
            try:
                result = suite.run(settings)
            finally:
                text.MAX_ERROR_LINES = old_max_error_lines
                text.MAX_ASSIGN_LENGTH = old_max_assign_length
            LOGGER.info("Tests execution ended. Statistics:\n%s"
                        % result.suite.stat_message)
            if settings.log or settings.report or settings.xunit:
                writer = ResultWriter(settings.output if settings.log
                                      else result)
                writer.write_results(settings.get_rebot_settings())
        return result.return_code

    def validate(self, options, arguments):
        return self._filter_options_without_value(options), arguments

    def _filter_options_without_value(self, options):
        return dict((name, value) for name, value in options.items()
                    if value not in (None, []))


def run_cli(arguments=None, exit=True):
    """Command line execution entry point for running tests.

    :param arguments: Command line options and arguments as a list of strings.
        Defaults to ``sys.argv[1:]`` if not given.
    :param exit: If ``True``, call ``sys.exit`` with the return code denoting
        execution status, otherwise just return the rc.

    Entry point used when running tests from the command line, but can also
    be used by custom scripts that execute tests. Especially useful if the
    script itself needs to accept same arguments as accepted by Robot Framework,
    because the script can just pass them forward directly along with the
    possible default values it sets itself.

    Example::

        from robot import run_cli

        # Run tests and return the return code.
        rc = run_cli(['--name', 'Example', 'tests.robot'], exit=False)

        # Run tests and exit to the system automatically.
        run_cli(['--name', 'Example', 'tests.robot'])

    See also the :func:`run` function that allows setting options as keyword
    arguments like ``name="Example"`` and generally has a richer API for
    programmatic test execution.
    """
    if arguments is None:
        arguments = sys.argv[1:]
    return RobotFramework().execute_cli(arguments, exit=exit)


def run(*tests, **options):
    """Programmatic entry point for running tests.

    :param tests: Paths to test case files/directories to be executed similarly
        as when running the ``robot`` command on the command line.
    :param options: Options to configure and control execution. Accepted
        options are mostly same as normal command line options to the ``robot``
        command. Option names match command line option long names without
        hyphens so that, for example, ``--name`` becomes ``name``.

    Most options that can be given from the command line work. An exception
    is that options ``--pythonpath``, ``--argumentfile``, ``--help`` and
    ``--version`` are not supported.

    Options that can be given on the command line multiple times can be
    passed as lists. For example, ``include=['tag1', 'tag2']`` is equivalent
    to ``--include tag1 --include tag2``. If such options are used only once,
    they can be given also as a single string like ``include='tag'``.

    Options that accept no value can be given as Booleans. For example,
    ``dryrun=True`` is same as using the ``--dryrun`` option.

    Options that accept string ``NONE`` as a special value can also be used
    with Python ``None``. For example, using ``log=None`` is equivalent to
    ``--log NONE``.

    ``listener``, ``prerunmodifier`` and ``prerebotmodifier`` options allow
    passing values as Python objects in addition to module names these command
    line options support. For example, ``run('tests', listener=MyListener())``.

    To capture the standard output and error streams, pass an open file or
    file-like object as special keyword arguments ``stdout`` and ``stderr``,
    respectively.

    A return code is returned similarly as when running on the command line.
    Zero means that tests were executed and no test failed, values up to 250
    denote the number of failed tests, and values between 251-255 are for other
    statuses documented in the Robot Framework User Guide.

    Example::

        from robot import run

        run('path/to/tests.robot')
        run('tests.robot', include=['tag1', 'tag2'], splitlog=True)
        with open('stdout.txt', 'w') as stdout:
            run('t1.robot', 't2.robot', name='Example', log=None, stdout=stdout)

    Equivalent command line usage::

        robot path/to/tests.robot
        robot --include tag1 --include tag2 --splitlog tests.robot
        robot --name Example --log NONE t1.robot t2.robot > stdout.txt
    """
    return RobotFramework().execute(*tests, **options)


if __name__ == '__main__':
    run_cli(sys.argv[1:])



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from copy import copy
import inspect

from robot.utils import (getdoc, getshortdoc, is_list_like, normpath, printable_name,
                         split_tags_from_doc, type_name)
from robot.errors import DataError
from robot.model import Tags

from .arguments import ArgumentSpec, DynamicArgumentParser, PythonArgumentParser
from .dynamicmethods import GetKeywordSource, GetKeywordTypes
from .librarykeywordrunner import (EmbeddedArgumentsRunner,
                                   LibraryKeywordRunner, RunKeywordRunner)
from .runkwregister import RUN_KW_REGISTER


def Handler(library, name, method):
    if RUN_KW_REGISTER.is_run_keyword(library.orig_name, name):
        return _RunKeywordHandler(library, name, method)
    return _PythonHandler(library, name, method)


def DynamicHandler(library, name, method, doc, argspec, tags=None):
    if RUN_KW_REGISTER.is_run_keyword(library.orig_name, name):
        return _DynamicRunKeywordHandler(library, name, method, doc, argspec, tags)
    return _DynamicHandler(library, name, method, doc, argspec, tags)


def InitHandler(library, method=None, docgetter=None):
    return _PythonInitHandler(library, '__init__', method, docgetter)


class _RunnableHandler:
    supports_embedded_args = False

    def __init__(self, library, handler_name, handler_method, doc='', tags=None):
        self.library = library
        self._handler_name = handler_name
        self.name = self._get_name(handler_name, handler_method)
        self.arguments = self._parse_arguments(handler_method)
        self._method = self._get_initial_handler(library, handler_name,
                                                 handler_method)
        doc, tags_from_doc = split_tags_from_doc(doc or '')
        tags_from_attr = self._get_tags_from_attribute(handler_method)
        self._doc = doc
        self.tags = Tags(tuple(tags_from_doc) +
                         tuple(tags_from_attr) +
                         tuple(tags or ()))

    def _get_name(self, handler_name, handler_method):
        robot_name = getattr(handler_method, 'robot_name', None)
        name = robot_name or printable_name(handler_name, code_style=True)
        if not name:
            raise DataError('Keyword name cannot be empty.')
        return name

    def _parse_arguments(self, handler_method):
        raise NotImplementedError

    def _get_tags_from_attribute(self, handler_method):
        tags = getattr(handler_method, 'robot_tags', ())
        if not is_list_like(tags):
            raise DataError(f"Expected tags to be list-like, got {type_name(tags)}.")
        return tags

    def _get_initial_handler(self, library, name, method):
        if library.scope.is_global:
            return self._get_global_handler(method, name)
        return None

    def resolve_arguments(self, args, variables=None, languages=None):
        return self.arguments.resolve(args, variables, self.library.converters,
                                      languages=languages)

    @property
    def doc(self):
        return self._doc

    @property
    def longname(self):
        return f'{self.library.name}.{self.name}'

    @property
    def shortdoc(self):
        return getshortdoc(self.doc)

    @property
    def libname(self):
        return self.library.name

    @property
    def source(self):
        return self.library.source

    @property
    def lineno(self):
        return -1

    def create_runner(self, name, languages=None):
        return LibraryKeywordRunner(self, languages=languages)

    def current_handler(self):
        if self._method:
            return self._method
        return self._get_handler(self.library.get_instance(), self._handler_name)

    def _get_global_handler(self, method, name):
        return method

    def _get_handler(self, lib_instance, handler_name):
        try:
            return getattr(lib_instance, handler_name)
        except AttributeError:
            # Occurs with old-style classes.
            if handler_name == '__init__':
                return None
            raise


class _PythonHandler(_RunnableHandler):

    def __init__(self, library, handler_name, handler_method):
        super().__init__(library, handler_name, handler_method, getdoc(handler_method))

    def _parse_arguments(self, handler_method):
        return PythonArgumentParser().parse(handler_method, self.longname)

    @property
    def source(self):
        handler = self.current_handler()
        # `getsourcefile` can return None and raise TypeError.
        try:
            source = inspect.getsourcefile(inspect.unwrap(handler))
        except TypeError:
            source = None
        return normpath(source) if source else self.library.source

    @property
    def lineno(self):
        handler = self.current_handler()
        try:
            lines, start_lineno = inspect.getsourcelines(inspect.unwrap(handler))
        except (TypeError, OSError, IOError):
            return -1
        for increment, line in enumerate(lines):
            if line.strip().startswith('def '):
                return start_lineno + increment
        return start_lineno


class _DynamicHandler(_RunnableHandler):

    def __init__(self, library, handler_name, dynamic_method, doc='', argspec=None,
                 tags=None):
        self._argspec = argspec
        self._run_keyword_method_name = dynamic_method.name
        self._supports_kwargs = dynamic_method.supports_kwargs
        # Cannot use super() here due to multi-inheritance in _DynamicRunKeywordHandler
        _RunnableHandler.__init__(self, library, handler_name, dynamic_method.method,
                                  doc, tags)
        self._source_info = None

    def _parse_arguments(self, handler_method):
        spec = DynamicArgumentParser().parse(self._argspec, self.longname)
        if not self._supports_kwargs:
            name = self._run_keyword_method_name
            if spec.var_named:
                raise DataError(f"Too few '{name}' method parameters for "
                                f"**kwargs support.")
            if spec.named_only:
                raise DataError(f"Too few '{name}' method parameters for "
                                f"keyword-only arguments support.")
        get_keyword_types = GetKeywordTypes(self.library.get_instance())
        spec.types = get_keyword_types(self._handler_name)
        return spec

    @property
    def source(self):
        if self._source_info is None:
            self._source_info = self._get_source_info()
        return self._source_info[0]

    def _get_source_info(self):
        get_keyword_source = GetKeywordSource(self.library.get_instance())
        try:
            source = get_keyword_source(self._handler_name)
        except DataError as err:
            self.library.report_error(
                f"Getting source information for keyword '{self.name}' failed: {err}",
                err.details
            )
            source = None
        if source and ':' in source and source.rsplit(':', 1)[1].isdigit():
            source, lineno = source.rsplit(':', 1)
            lineno = int(lineno)
        else:
            lineno = -1
        return normpath(source) if source else self.library.source, lineno

    @property
    def lineno(self):
        if self._source_info is None:
            self._source_info = self._get_source_info()
        return self._source_info[1]

    def resolve_arguments(self, arguments, variables=None, languages=None):
        positional, named = super().resolve_arguments(arguments, variables, languages)
        if not self._supports_kwargs:
            positional, named = self.arguments.map(positional, named)
        return positional, named

    def _get_handler(self, lib_instance, handler_name):
        runner = getattr(lib_instance, self._run_keyword_method_name)
        return self._get_dynamic_handler(runner, handler_name)

    def _get_global_handler(self, method, name):
        return self._get_dynamic_handler(method, name)

    def _get_dynamic_handler(self, runner, name):
        def handler(*positional, **kwargs):
            if self._supports_kwargs:
                return runner(name, positional, kwargs)
            else:
                return runner(name, positional)
        return handler


class _RunKeywordHandler(_PythonHandler):

    def create_runner(self, name, languages=None):
        dry_run = RUN_KW_REGISTER.get_dry_run(self.library.orig_name, self.name)
        return RunKeywordRunner(self, execute_in_dry_run=dry_run)

    @property
    def _args_to_process(self):
        return RUN_KW_REGISTER.get_args_to_process(self.library.orig_name, self.name)

    def resolve_arguments(self, args, variables=None, languages=None):
        return self.arguments.resolve(args, variables, self.library.converters,
                                      resolve_named=False,
                                      resolve_variables_until=self._args_to_process)


class _DynamicRunKeywordHandler(_DynamicHandler, _RunKeywordHandler):
    _parse_arguments = _RunKeywordHandler._parse_arguments
    resolve_arguments = _RunKeywordHandler.resolve_arguments


class _PythonInitHandler(_PythonHandler):

    def __init__(self, library, handler_name, handler_method, docgetter):
        super().__init__(library, handler_name, handler_method)
        self._docgetter = docgetter

    def _get_name(self, handler_name, handler_method):
        return '__init__'

    @property
    def doc(self):
        if self._docgetter:
            self._doc = self._docgetter() or self._doc
            self._docgetter = None
        return self._doc

    def _parse_arguments(self, init_method):
        parser = PythonArgumentParser(type='Library')
        return parser.parse(init_method or (lambda: None), self.library.name)


class EmbeddedArgumentsHandler:
    supports_embedded_args = True

    def __init__(self, embedded, orig_handler):
        self.arguments = ArgumentSpec()  # Show empty argument spec for Libdoc
        self.embedded = embedded
        self._orig_handler = orig_handler

    def __getattr__(self, item):
        return getattr(self._orig_handler, item)

    @property
    def library(self):
        return self._orig_handler.library

    @library.setter
    def library(self, library):
        self._orig_handler.library = library

    def matches(self, name):
        return self.embedded.match(name) is not None

    def create_runner(self, name, languages=None):
        return EmbeddedArgumentsRunner(self, name)

    def resolve_arguments(self, args, variables=None, languages=None):
        argspec = self._orig_handler.arguments
        if variables:
            if argspec.var_positional:
                args = variables.replace_list(args)
            else:
                args = [variables.replace_scalar(a) for a in args]
            self.embedded.validate(args)
        return argspec.convert(args, named={}, converters=self.library.converters,
                               dry_run=not variables)

    def __copy__(self):
        return EmbeddedArgumentsHandler(self.embedded, copy(self._orig_handler))



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Exceptions that libraries can use for communicating failures and other events.

These exceptions can be imported also via the top level :mod:`robot.api` package like
``from robot.api import SkipExecution``.

This module and all exceptions are new in Robot Framework 4.0.
"""


class Failure(AssertionError):
    """Report failed validation.

    There is no practical difference in using this exception compared to using
    the standard ``AssertionError``. The main benefits are HTML support and that
    the name of this exception is consistent with other exceptions in this module.
    """
    ROBOT_SUPPRESS_NAME = True

    def __init__(self, message, html=False):
        """
        :param message: Exception message.
        :param html: When ``True``, message is considered to be HTML and not escaped.
        """
        AssertionError.__init__(self, message if not html else '*HTML* ' + message)


class ContinuableFailure(Failure):
    """Report failed validation but allow continuing execution."""
    ROBOT_CONTINUE_ON_FAILURE = True


class Error(RuntimeError):
    """Report error in execution.

    Failures related to the system not behaving as expected should typically be
    reported using the :class:`Failure` exception or the standard ``AssertionError``.
    This exception can be used, for example, if the keyword is used incorrectly.

    There is no practical difference in using this exception compared to using
    the standard ``RuntimeError``. The main benefits are HTML support and that
    the name of this exception is consistent with other exceptions in this module.
    """
    ROBOT_SUPPRESS_NAME = True

    def __init__(self, message, html=False):
        """
        :param message: Exception message.
        :param html: When ``True``, message is considered to be HTML and not escaped.
        """
        RuntimeError.__init__(self, message if not html else '*HTML* ' + message)


class FatalError(Error):
    """Report error that stops the whole execution."""
    ROBOT_EXIT_ON_FAILURE = True
    ROBOT_SUPPRESS_NAME = False


class SkipExecution(Exception):
    """Mark the executed test or task skipped."""
    ROBOT_SKIP_EXECUTION = True
    ROBOT_SUPPRESS_NAME = True

    def __init__(self, message, html=False):
        """
        :param message: Exception message.
        :param html: When ``True``, message is considered to be HTML and not escaped.
        """
        Exception.__init__(self, message if not html else '*HTML* ' + message)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
from typing import (Any, Callable, cast, Generic, Iterable, Type, TYPE_CHECKING,
                    TypeVar, Union)

from robot.utils import copy_signature, KnownAtRuntime

from .itemlist import ItemList
from .modelobject import DataDict, full_name, ModelObject

if TYPE_CHECKING:
    from robot.result.model import ForIteration, WhileIteration
    from robot.running.model import UserKeyword, ResourceFile
    from .control import (Break, Continue, Error, For, If, IfBranch, Return,
                          Try, TryBranch, While)
    from .keyword import Keyword
    from .message import Message
    from .testcase import TestCase
    from .testsuite import TestSuite


BodyItemParent = Union['TestSuite', 'TestCase', 'UserKeyword', 'For', 'ForIteration',
                       'If', 'IfBranch', 'Try', 'TryBranch', 'While', 'WhileIteration',
                       'Keyword', 'Return', 'Continue', 'Break', 'Error', None]
BI = TypeVar('BI', bound='BodyItem')
KW = TypeVar('KW', bound='Keyword')
F = TypeVar('F', bound='For')
W = TypeVar('W', bound='While')
I = TypeVar('I', bound='If')
T = TypeVar('T', bound='Try')
R = TypeVar('R', bound='Return')
C = TypeVar('C', bound='Continue')
B = TypeVar('B', bound='Break')
M = TypeVar('M', bound='Message')
E = TypeVar('E', bound='Error')
IT = TypeVar('IT', bound='IfBranch|TryBranch')


class BodyItem(ModelObject):
    KEYWORD = 'KEYWORD'
    SETUP = 'SETUP'
    TEARDOWN = 'TEARDOWN'
    FOR = 'FOR'
    ITERATION = 'ITERATION'
    IF_ELSE_ROOT = 'IF/ELSE ROOT'
    IF = 'IF'
    ELSE_IF = 'ELSE IF'
    ELSE = 'ELSE'
    TRY_EXCEPT_ROOT = 'TRY/EXCEPT ROOT'
    TRY = 'TRY'
    EXCEPT = 'EXCEPT'
    FINALLY = 'FINALLY'
    WHILE = 'WHILE'
    RETURN = 'RETURN'
    CONTINUE = 'CONTINUE'
    BREAK = 'BREAK'
    ERROR = 'ERROR'
    MESSAGE = 'MESSAGE'
    type = None
    __slots__ = ['parent']

    @property
    def id(self) -> 'str|None':
        """Item id in format like ``s1-t3-k1``.

        See :attr:`TestSuite.id <robot.model.testsuite.TestSuite.id>` for
        more information.

        ``id`` is ``None`` only in these special cases:

        - Keyword uses a placeholder for ``setup`` or ``teardown`` when
          a ``setup`` or ``teardown`` is not actually used.
        - With :class:`~robot.model.control.If` and :class:`~robot.model.control.Try`
          instances representing IF/TRY structure roots.
        """
        return self._get_id(self.parent)

    def _get_id(self, parent: 'BodyItemParent|ResourceFile') -> str:
        if not parent:
            return 'k1'
        # This algorithm must match the id creation algorithm in the JavaScript side
        # or linking to warnings and errors won't work.
        steps = []
        if getattr(parent, 'has_setup', False):
            steps.append(parent.setup)          # type: ignore - Use Protocol with RF 7.
        if hasattr(parent, 'body'):
            steps.extend(step for step in
                         parent.body.flatten()  # type: ignore - Use Protocol with RF 7.
                         if step.type != self.MESSAGE)
        if getattr(parent, 'has_teardown', False):
            steps.append(parent.teardown)       # type: ignore - Use Protocol with RF 7.
        index = steps.index(self) if self in steps else len(steps)
        parent_id = getattr(parent, 'id', None)
        return f'{parent_id}-k{index + 1}' if parent_id else f'k{index + 1}'

    def to_dict(self) -> DataDict:
        raise NotImplementedError


class BaseBody(ItemList[BodyItem], Generic[KW, F, W, I, T, R, C, B, M, E]):
    """Base class for Body and Branches objects."""
    __slots__ = []
    # Set using 'BaseBody.register' when these classes are created.
    keyword_class: Type[KW] = KnownAtRuntime
    for_class: Type[F] = KnownAtRuntime
    while_class: Type[W] = KnownAtRuntime
    if_class: Type[I] = KnownAtRuntime
    try_class: Type[T] = KnownAtRuntime
    return_class: Type[R] = KnownAtRuntime
    continue_class: Type[C] = KnownAtRuntime
    break_class: Type[B] = KnownAtRuntime
    message_class: Type[M] = KnownAtRuntime
    error_class: Type[E] = KnownAtRuntime

    def __init__(self, parent: BodyItemParent = None,
                 items: 'Iterable[BodyItem|DataDict]' = ()):
        super().__init__(BodyItem, {'parent': parent}, items)

    def _item_from_dict(self, data: DataDict) -> BodyItem:
        item_type = data.get('type', None)
        if not item_type:
            item_class = self.keyword_class
        elif item_type == BodyItem.IF_ELSE_ROOT:
            item_class = self.if_class
        elif item_type == BodyItem.TRY_EXCEPT_ROOT:
            item_class = self.try_class
        else:
            item_class = getattr(self, item_type.lower() + '_class')
        item_class = cast(Type[BodyItem], item_class)
        return item_class.from_dict(data)

    @classmethod
    def register(cls, item_class: Type[BI]) -> Type[BI]:
        name_parts = re.findall('([A-Z][a-z]+)', item_class.__name__) + ['class']
        name = '_'.join(name_parts).lower()
        if not hasattr(cls, name):
            raise TypeError(f"Cannot register '{name}'.")
        setattr(cls, name, item_class)
        return item_class

    @property
    def create(self):
        raise AttributeError(
            f"'{full_name(self)}' object has no attribute 'create'. "
            f"Use item specific methods like 'create_keyword' instead."
        )

    def _create(self, cls: 'Type[BI]', name: str, args: 'tuple[Any]',
                kwargs: 'dict[str, Any]') -> BI:
        if cls is KnownAtRuntime:
            raise TypeError(f"'{full_name(self)}' object does not support '{name}'.")
        return self.append(cls(*args, **kwargs))  # type: ignore

    @copy_signature(keyword_class)
    def create_keyword(self, *args, **kwargs) -> keyword_class:
        return self._create(self.keyword_class, 'create_keyword', args, kwargs)

    @copy_signature(for_class)
    def create_for(self, *args, **kwargs) -> for_class:
        return self._create(self.for_class, 'create_for', args, kwargs)

    @copy_signature(if_class)
    def create_if(self, *args, **kwargs) -> if_class:
        return self._create(self.if_class, 'create_if', args, kwargs)

    @copy_signature(try_class)
    def create_try(self, *args, **kwargs) -> try_class:
        return self._create(self.try_class, 'create_try', args, kwargs)

    @copy_signature(while_class)
    def create_while(self, *args, **kwargs) -> while_class:
        return self._create(self.while_class, 'create_while', args, kwargs)

    @copy_signature(return_class)
    def create_return(self, *args, **kwargs) -> return_class:
        return self._create(self.return_class, 'create_return', args, kwargs)

    @copy_signature(continue_class)
    def create_continue(self, *args, **kwargs) -> continue_class:
        return self._create(self.continue_class, 'create_continue', args, kwargs)

    @copy_signature(break_class)
    def create_break(self, *args, **kwargs) -> break_class:
        return self._create(self.break_class, 'create_break', args, kwargs)

    @copy_signature(message_class)
    def create_message(self, *args, **kwargs) -> message_class:
        return self._create(self.message_class, 'create_message', args, kwargs)

    @copy_signature(error_class)
    def create_error(self, *args, **kwargs) -> error_class:
        return self._create(self.error_class, 'create_error', args, kwargs)

    def filter(self, keywords: 'bool|None' = None, messages: 'bool|None' = None,
               predicate: 'Callable[[T], bool]|None' = None):
        """Filter body items based on type and/or custom predicate.

        To include or exclude items based on types, give matching arguments
        ``True`` or ``False`` values. For example, to include only keywords,
        use ``body.filter(keywords=True)`` and to exclude messages use
        ``body.filter(messages=False)``. Including and excluding by types
        at the same time is not supported and filtering my ``messages``
        is supported only if the ``Body`` object actually supports messages.

        Custom ``predicate`` is a callable getting each body item as an argument
        that must return ``True/False`` depending on should the item be included
        or not.

        Selected items are returned as a list and the original body is not modified.

        It was earlier possible to filter also based on FOR and IF types.
        That support was removed in RF 5.0 because it was not considered useful
        in general and because adding support for all new control structures
        would have required extra work. To exclude all control structures, use
        ``body.filter(keywords=True, messages=True)`` and to only include them
        use ``body.filter(keywords=False``, messages=False)``. For more detailed
        filtering it is possible to use ``predicate``.
        """
        if messages is not None and self.message_class is KnownAtRuntime:
            raise TypeError(f"'{full_name(self)}' object does not support "
                            f"filtering by 'messages'.")
        return self._filter([(self.keyword_class, keywords),
                             (self.message_class, messages)], predicate)

    def _filter(self, types, predicate):
        include = tuple(cls for cls, activated in types if activated is True and cls)
        exclude = tuple(cls for cls, activated in types if activated is False and cls)
        if include and exclude:
            raise ValueError('Items cannot be both included and excluded by type.')
        items = list(self)
        if include:
            items = [item for item in items if isinstance(item, include)]
        if exclude:
            items = [item for item in items if not isinstance(item, exclude)]
        if predicate:
            items = [item for item in items if predicate(item)]
        return items

    def flatten(self) -> 'list[BodyItem]':
        """Return steps so that IF and TRY structures are flattened.

        Basically the IF/ELSE and TRY/EXCEPT root elements are replaced
        with their branches. This is how they are shown in log files.
        """
        roots = BodyItem.IF_ELSE_ROOT, BodyItem.TRY_EXCEPT_ROOT
        steps = []
        for item in self:
            if item.type in roots:
                item = cast('Try|If', item)
                steps.extend(item.body)
            else:
                steps.append(item)
        return steps


class Body(BaseBody['Keyword', 'For', 'While', 'If', 'Try', 'Return', 'Continue',
                    'Break', 'Message', 'Error']):
    """A list-like object representing a body of a test, keyword, etc.

    Body contains the keywords and other structures such as FOR loops.
    """
    pass


class BranchType(Generic[IT]):
    """Class that wrapps `Generic` as python doesn't allow multple generic inheritance"""
    pass


class BaseBranches(BaseBody[KW, F, W, I, T, R, C, B, M, E], BranchType[IT]):
    """A list-like object representing IF and TRY branches."""
    __slots__ = ['branch_class']
    branch_type: Type[IT] = KnownAtRuntime

    def __init__(self, branch_class: Type[IT],
                 parent: BodyItemParent = None,
                 items: 'Iterable[BodyItem|DataDict]' = ()):
        self.branch_class = branch_class
        super().__init__(parent, items)

    def _item_from_dict(self, data: DataDict) -> IT:
        return self.branch_class.from_dict(data)

    @copy_signature(branch_type)
    def create_branch(self, *args, **kwargs) -> IT:
        return self._create(self.branch_class, 'create_branch', args, kwargs)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

class JsonWriter:

    def __init__(self, output, separator=''):
        self._writer = JsonDumper(output)
        self._separator = separator

    def write_json(self, prefix, data, postfix=';\n', mapping=None,
                   separator=True):
        self._writer.write(prefix)
        self._writer.dump(data, mapping)
        self._writer.write(postfix)
        self._write_separator(separator)

    def write(self, string, postfix=';\n', separator=True):
        self._writer.write(string + postfix)
        self._write_separator(separator)

    def _write_separator(self, separator):
        if separator and self._separator:
            self._writer.write(self._separator)


class JsonDumper:

    def __init__(self, output):
        self.write = output.write
        self._dumpers = (MappingDumper(self),
                         IntegerDumper(self),
                         TupleListDumper(self),
                         StringDumper(self),
                         NoneDumper(self),
                         DictDumper(self))

    def dump(self, data, mapping=None):
        for dumper in self._dumpers:
            if dumper.handles(data, mapping):
                dumper.dump(data, mapping)
                return
        raise ValueError('Dumping %s not supported.' % type(data))


class _Dumper:
    _handled_types = None

    def __init__(self, jsondumper):
        self._dump = jsondumper.dump
        self._write = jsondumper.write

    def handles(self, data, mapping):
        return isinstance(data, self._handled_types)

    def dump(self, data, mapping):
        raise NotImplementedError


class StringDumper(_Dumper):
    _handled_types = str
    _search_and_replace = [('\\', '\\\\'), ('"', '\\"'), ('\t', '\\t'),
                           ('\n', '\\n'), ('\r', '\\r'), ('</', '\\x3c/')]

    def dump(self, data, mapping):
        self._write('"%s"' % (self._escape(data) if data else ''))

    def _escape(self, string):
        for search, replace in self._search_and_replace:
            if search in string:
                string = string.replace(search, replace)
        return string


class IntegerDumper(_Dumper):
    # Handles also bool
    _handled_types = int

    def dump(self, data, mapping):
        self._write(str(data).lower())


class DictDumper(_Dumper):
    _handled_types = dict

    def dump(self, data, mapping):
        write = self._write
        dump = self._dump
        write('{')
        last_index = len(data) - 1
        for index, key in enumerate(sorted(data)):
            dump(key, mapping)
            write(':')
            dump(data[key], mapping)
            if index < last_index:
                write(',')
        write('}')


class TupleListDumper(_Dumper):
    _handled_types = (tuple, list)

    def dump(self, data, mapping):
        write = self._write
        dump = self._dump
        write('[')
        last_index = len(data) - 1
        for index, item in enumerate(data):
            dump(item, mapping)
            if index < last_index:
                write(',')
        write(']')


class MappingDumper(_Dumper):

    def handles(self, data, mapping):
        try:
            return mapping and data in mapping
        except TypeError:
            return False

    def dump(self, data, mapping):
        self._write(mapping[data])


class NoneDumper(_Dumper):

    def handles(self, data, mapping):
        return data is None

    def dump(self, data, mapping):
        self._write('null')



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from io import StringIO
import sys

from robot.output import LOGGER
from robot.utils import console_decode, console_encode


class OutputCapturer:

    def __init__(self, library_import=False):
        self._library_import = library_import
        self._python_out = PythonCapturer(stdout=True)
        self._python_err = PythonCapturer(stdout=False)

    def __enter__(self):
        if self._library_import:
            LOGGER.enable_library_import_logging()
        return self

    def __exit__(self, exc_type, exc_value, exc_trace):
        self._release_and_log()
        if self._library_import:
            LOGGER.disable_library_import_logging()
        return False

    def _release_and_log(self):
        stdout, stderr = self._release()
        if stdout:
            LOGGER.log_output(stdout)
        if stderr:
            LOGGER.log_output(stderr)
            sys.__stderr__.write(console_encode(stderr, stream=sys.__stderr__))

    def _release(self):
        stdout = self._python_out.release()
        stderr = self._python_err.release()
        return stdout, stderr


class PythonCapturer:

    def __init__(self, stdout=True):
        if stdout:
            self._original = sys.stdout
            self._set_stream = self._set_stdout
        else:
            self._original = sys.stderr
            self._set_stream = self._set_stderr
        self._stream = StringIO()
        self._set_stream(self._stream)

    def _set_stdout(self, stream):
        sys.stdout = stream

    def _set_stderr(self, stream):
        sys.stderr = stream

    def release(self):
        # Original stream must be restored before closing the current
        self._set_stream(self._original)
        try:
            return self._get_value(self._stream)
        finally:
            self._stream.close()
            self._avoid_at_exit_errors(self._stream)

    def _get_value(self, stream):
        try:
            return console_decode(stream.getvalue())
        except UnicodeError:
            # Error occurs if non-ASCII chars logged both as str and unicode.
            stream.buf = console_decode(stream.buf)
            stream.buflist = [console_decode(item) for item in stream.buflist]
            return stream.getvalue()

    def _avoid_at_exit_errors(self, stream):
        # Avoid ValueError at program exit when logging module tries to call
        # methods of streams it has intercepted that are already closed.
        # Which methods are called, and does logging silence possible errors,
        # depends on Python version. For related discussion see
        # http://bugs.python.org/issue6333
        stream.write = lambda s: None
        stream.flush = lambda: None



#!/usr/bin/env python

#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Module implementing the command line entry point for the `Testdoc` tool.

This module can be executed from the command line using the following
approaches::

    python -m robot.testdoc
    python path/to/robot/testdoc.py

Instead of ``python`` it is possible to use also other Python interpreters.

This module also provides :func:`testdoc` and :func:`testdoc_cli` functions
that can be used programmatically. Other code is for internal usage.
"""

import sys
import time
from pathlib import Path

if __name__ == '__main__' and 'robot' not in sys.modules:
    import pythonpathsetter

from robot.conf import RobotSettings
from robot.htmldata import HtmlFileWriter, ModelWriter, JsonWriter, TESTDOC
from robot.running import TestSuiteBuilder
from robot.utils import (abspath, Application, file_writer, get_link_path,
                         html_escape, html_format, is_list_like, secs_to_timestr,
                         seq2str2, timestr_to_secs, unescape)


USAGE = """robot.testdoc -- Robot Framework test data documentation tool

Version:  <VERSION>

Usage:  python -m robot.testdoc [options] data_sources output_file

Testdoc generates a high level test documentation based on Robot Framework
test data. Generated documentation includes name, documentation and other
metadata of each test suite and test case, as well as the top-level keywords
and their arguments.

Options
=======

  -T --title title       Set the title of the generated documentation.
                         Underscores in the title are converted to spaces.
                         The default title is the name of the top level suite.
  -N --name name         Override the name of the top level suite.
  -D --doc document      Override the documentation of the top level suite.
  -M --metadata name:value *  Set/override metadata of the top level suite.
  -G --settag tag *      Set given tag(s) to all test cases.
  -t --test name *       Include tests by name.
  -s --suite name *      Include suites by name.
  -i --include tag *     Include tests by tags.
  -e --exclude tag *     Exclude tests by tags.
  -A --argumentfile path *  Text file to read more arguments from. Use special
                          path `STDIN` to read contents from the standard input
                          stream. File can have both options and data sources
                          one per line. Contents do not need to be escaped but
                          spaces in the beginning and end of lines are removed.
                          Empty lines and lines starting with a hash character
                          (#) are ignored.
                          Example file:
                          |  --name Example
                          |  # This is a comment line
                          |  my_tests.robot
                          |  output.html
                          Examples:
                          --argumentfile argfile.txt --argumentfile STDIN
  -h -? --help           Print this help.

All options except --title have exactly same semantics as same options have
when executing test cases.

Execution
=========

Data can be given as a single file, directory, or as multiple files and
directories. In all these cases, the last argument must be the file where
to write the output. The output is always created in HTML format.

Testdoc works with all interpreters supported by Robot Framework.
It can be executed as an installed module like
`python -m robot.testdoc` or as a script like `python path/robot/testdoc.py`.

Examples:

  python -m robot.testdoc my_test.robot testdoc.html
  python path/to/robot/testdoc.py first_suite.txt second_suite.txt output.html

For more information about Testdoc and other built-in tools, see
http://robotframework.org/robotframework/#built-in-tools.
"""


class TestDoc(Application):

    def __init__(self):
        Application.__init__(self, USAGE, arg_limits=(2,))

    def main(self, datasources, title=None, **options):
        outfile = abspath(datasources.pop())
        suite = TestSuiteFactory(datasources, **options)
        self._write_test_doc(suite, outfile, title)
        self.console(outfile)

    def _write_test_doc(self, suite, outfile, title):
        with file_writer(outfile, usage='Testdoc output') as output:
            model_writer = TestdocModelWriter(output, suite, title)
            HtmlFileWriter(output, model_writer).write(TESTDOC)


def TestSuiteFactory(datasources, **options):
    settings = RobotSettings(options)
    if not is_list_like(datasources):
        datasources = [datasources]
    suite = TestSuiteBuilder(process_curdir=False).build(*datasources)
    suite.configure(**settings.suite_config)
    return suite


class TestdocModelWriter(ModelWriter):

    def __init__(self, output, suite, title=None):
        self._output = output
        self._output_path = getattr(output, 'name', None)
        self._suite = suite
        self._title = title.replace('_', ' ') if title else suite.name

    def write(self, line):
        self._output.write('<script type="text/javascript">\n')
        self.write_data()
        self._output.write('</script>\n')

    def write_data(self):
        model = {
            'suite': JsonConverter(self._output_path).convert(self._suite),
            'title': self._title,
            'generated': int(time.time() * 1000)
        }
        JsonWriter(self._output).write_json('testdoc = ', model)


class JsonConverter:

    def __init__(self, output_path=None):
        self._output_path = output_path

    def convert(self, suite):
        return self._convert_suite(suite)

    def _convert_suite(self, suite):
        return {
            'source': str(suite.source or ''),
            'relativeSource': self._get_relative_source(suite.source),
            'id': suite.id,
            'name': self._escape(suite.name),
            'fullName': self._escape(suite.longname),
            'doc': self._html(suite.doc),
            'metadata': [(self._escape(name), self._html(value))
                         for name, value in suite.metadata.items()],
            'numberOfTests': suite.test_count   ,
            'suites': self._convert_suites(suite),
            'tests': self._convert_tests(suite),
            'keywords': list(self._convert_keywords((suite.setup, suite.teardown)))
        }

    def _get_relative_source(self, source):
        if not source or not self._output_path:
            return ''
        return get_link_path(source, Path(self._output_path).parent)

    def _escape(self, item):
        return html_escape(item)

    def _html(self, item):
        return html_format(unescape(item))

    def _convert_suites(self, suite):
        return [self._convert_suite(s) for s in suite.suites]

    def _convert_tests(self, suite):
        return [self._convert_test(t) for t in suite.tests]

    def _convert_test(self, test):
        if test.setup:
            test.body.insert(0, test.setup)
        if test.teardown:
            test.body.append(test.teardown)
        return {
            'name': self._escape(test.name),
            'fullName': self._escape(test.longname),
            'id': test.id,
            'doc': self._html(test.doc),
            'tags': [self._escape(t) for t in test.tags],
            'timeout': self._get_timeout(test.timeout),
            'keywords': list(self._convert_keywords(test.body))
        }

    def _convert_keywords(self, keywords):
        for kw in keywords:
            if not kw:
                continue
            if kw.type == kw.SETUP:
                yield self._convert_keyword(kw, 'SETUP')
            elif kw.type == kw.TEARDOWN:
                yield self._convert_keyword(kw, 'TEARDOWN')
            elif kw.type == kw.FOR:
                yield self._convert_for(kw)
            elif kw.type == kw.WHILE:
                yield self._convert_while(kw)
            elif kw.type == kw.IF_ELSE_ROOT:
                yield from self._convert_if(kw)
            elif kw.type == kw.TRY_EXCEPT_ROOT:
                yield from self._convert_try(kw)
            else:
                yield self._convert_keyword(kw, 'KEYWORD')

    def _convert_for(self, data):
        name = '%s %s %s' % (', '.join(data.variables), data.flavor,
                             seq2str2(data.values))
        return {'type': 'FOR', 'name': self._escape(name), 'arguments': ''}

    def _convert_while(self, data):
        return {'type': 'WHILE', 'name': self._escape(data.condition), 'arguments': ''}

    def _convert_if(self, data):
        for branch in data.body:
            yield {'type': branch.type,
                   'name': self._escape(branch.condition or ''),
                   'arguments': ''}

    def _convert_try(self, data):
        for branch in data.body:
            if branch.type == branch.EXCEPT:
                patterns = ', '.join(branch.patterns)
                as_var = f'AS {branch.variable}' if branch.variable else ''
                name = f'{patterns} {as_var}'.strip()
            else:
                name = ''
            yield {'type': branch.type, 'name': name, 'arguments': ''}

    def _convert_keyword(self, kw, kw_type):
        return {
            'type': kw_type,
            'name': self._escape(self._get_kw_name(kw)),
            'arguments': self._escape(', '.join(kw.args))
        }

    def _get_kw_name(self, kw):
        if kw.assign:
            return '%s = %s' % (', '.join(a.rstrip('= ') for a in kw.assign), kw.name)
        return kw.name

    def _get_timeout(self, timeout):
        if timeout is None:
            return ''
        try:
            tout = secs_to_timestr(timestr_to_secs(timeout))
        except ValueError:
            tout = timeout
        return tout


def testdoc_cli(arguments):
    """Executes `Testdoc` similarly as from the command line.

    :param arguments: command line arguments as a list of strings.

    For programmatic usage the :func:`testdoc` function is typically better. It
    has a better API for that and does not call :func:`sys.exit` like
    this function.

    Example::

        from robot.testdoc import testdoc_cli

        testdoc_cli(['--title', 'Test Plan', 'mytests', 'plan.html'])
    """
    TestDoc().execute_cli(arguments)


def testdoc(*arguments, **options):
    """Executes `Testdoc` programmatically.

    Arguments and options have same semantics, and options have same names,
    as arguments and options to Testdoc.

    Example::

        from robot.testdoc import testdoc

        testdoc('mytests', 'plan.html', title='Test Plan')
    """
    TestDoc().execute(*arguments, **options)


if __name__ == '__main__':
    testdoc_cli(sys.argv[1:])



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

r"""Public API for parsing, inspecting and modifying test data.

Exposed API
-----------

The publicly exposed parsing entry points are the following:

* :func:`~.lexer.lexer.get_tokens`,
  :func:`~.lexer.lexer.get_resource_tokens`, and
  :func:`~.lexer.lexer.get_init_tokens`
  functions for `parsing data to tokens`_.

* :class:`~.lexer.tokens.Token` class that contains all token types as
  class attributes.

* :func:`~.parser.parser.get_model`,
  :func:`~.parser.parser.get_resource_model`, and
  :func:`~.parser.parser.get_init_model`
  functions for `parsing data to model`_ represented as
  an abstract syntax tree (AST).

* `Model objects`_ used by the AST model.

* :class:`~robot.parsing.model.visitor.ModelVisitor`
  to ease `inspecting model`_ and `modifying data`_.

* :class:`~robot.parsing.model.visitor.ModelTransformer`
  for `adding and removing nodes`_.

.. note:: This module is new in Robot Framework 4.0. In Robot Framework 3.2 functions
          for getting tokens and model as well as the :class:`~.lexer.tokens.Token`
          class were exposed directly via the :mod:`robot.api` package, but other
          parts of the parsing API were not publicly exposed. All code targeting
          Robot Framework 4.0 or newer should use this module because parsing related
          functions and classes will be removed from :mod:`robot.api` in the future.

.. note:: Parsing was totally rewritten in Robot Framework 3.2 and external
          tools using the parsing APIs need to be updated. Depending on
          the use case, it may be possible to use the higher level
          :func:`~robot.running.builder.builders.TestSuiteBuilder` instead.

Parsing data to tokens
----------------------

Data can be parsed to tokens by using
:func:`~.lexer.lexer.get_tokens`,
:func:`~.lexer.lexer.get_resource_tokens` or
:func:`~.lexer.lexer.get_init_tokens` functions depending on whether the data
represent a test case (or task) file, a resource file, or a suite
initialization file. In practice the difference between these functions is
what settings and sections are valid.

Typically the data is easier to inspect and modify by using the higher level
model discussed in the next section, but in some cases having just the tokens
can be enough. Tokens returned by the aforementioned functions are
:class:`~.lexer.tokens.Token` instances and they have the token type, value,
and position easily available as their attributes. Tokens also have useful
string representation used by the example below::

    from robot.api.parsing import get_tokens

    path = 'example.robot'

    for token in get_tokens(path):
        print(repr(token))

If the :file:`example.robot` used by the above example would contain

.. code-block:: robotframework

    *** Test Cases ***
    Example
        Keyword    argument

    Second example
        Keyword    xxx

    *** Keywords ***
    Keyword
        [Arguments]    ${arg}
        Log    ${arg}

then the beginning of the output got when running the earlier code would
look like this::

    Token(TESTCASE_HEADER, '*** Test Cases ***', 1, 0)
    Token(EOL, '\n', 1, 18)
    Token(EOS, '', 1, 19)
    Token(TESTCASE_NAME, 'Example', 2, 0)
    Token(EOL, '\n', 2, 7)
    Token(EOS, '', 2, 8)
    Token(SEPARATOR, '    ', 3, 0)
    Token(KEYWORD, 'Keyword', 3, 4)
    Token(SEPARATOR, '    ', 3, 11)
    Token(ARGUMENT, 'argument', 3, 15)
    Token(EOL, '\n', 3, 23)
    Token(EOS, '', 3, 24)
    Token(EOL, '\n', 4, 0)
    Token(EOS, '', 4, 1)

The output shows the token type, value, line number and column offset. When finding
tokens by their type, the constants in the :class:`~.lexer.tokens.Token` class such
as ``Token.TESTCASE_NAME`` and ``Token.EOL`` should be used instead the values
of these constants like ``'TESTCASE NAME'`` and ``'EOL'``. These values have
changed slightly in Robot Framework 4.0 and they may change in the future as well.

The ``EOL`` tokens denote end of a line and they include the newline character
and possible trailing spaces. The ``EOS`` tokens denote end of a logical
statement. Typically a single line forms a statement, but when the ``...``
syntax is used for continuation, a statement spans multiple lines. In
special cases a single line can also contain multiple statements.

Errors caused by unrecognized data such as non-existing section or setting names
are handled during the tokenizing phase. Such errors are reported using tokens
that have ``ERROR`` type and the actual error message in their ``error`` attribute.
Syntax errors such as empty FOR loops are only handled when building the higher
level model discussed below.

See the documentation of :func:`~.lexer.lexer.get_tokens` for details
about different ways how to specify the data to be parsed, how to control
should all tokens or only data tokens be returned, and should variables in
keyword arguments and elsewhere be tokenized or not.

Parsing data to model
---------------------

Data can be parsed to a higher level model by using
:func:`~.parser.parser.get_model`,
:func:`~.parser.parser.get_resource_model`, or
:func:`~.parser.parser.get_init_model` functions depending on the type of
the parsed file same way as when `parsing data to tokens`_.

The model is represented as an abstract syntax tree (AST) implemented on top
of Python's standard `ast.AST`_ class. To see how the model looks like, it is
possible to use the `ast.dump()`_ function or the third-party astpretty_
module::

    import ast
    import astpretty
    from robot.api.parsing import get_model

    model = get_model('example.robot')
    print(ast.dump(model, include_attributes=True))
    print('-' * 72)
    astpretty.pprint(model)

Running this code with the :file:`example.robot` file from the previous
section would produce so much output that it is not included here. If
you are going to work with Robot Framework's AST, you are recommended to
try that on your own.

.. _ast: https://docs.python.org/library/ast.html
.. _ast.AST: https://docs.python.org/library/ast.html#ast.AST
.. _ast.NodeVisitor: https://docs.python.org/library/ast.html#ast.NodeVisitor
.. _ast.NodeTransformer: https://docs.python.org/library/ast.html#ast.NodeTransformer
.. _ast.dump(): https://docs.python.org/library/ast.html#ast.dump
.. _astpretty: https://pypi.org/project/astpretty

Model objects
-------------

The model is build from nodes that are based `ast.AST`_ and further categorized
to blocks and statements. Blocks can contain other blocks and statements as
child nodes whereas statements only have tokens containing the actual data as
:class:`~.lexer.tokens.Token` instances. Both statements and blocks expose
their position information via ``lineno``, ``col_offset``, ``end_lineno`` and
``end_col_offset`` attributes and some nodes have also other special attributes
available.

Blocks:

- :class:`~robot.parsing.model.blocks.File` (the root of the model)
- :class:`~robot.parsing.model.blocks.SettingSection`
- :class:`~robot.parsing.model.blocks.VariableSection`
- :class:`~robot.parsing.model.blocks.TestCaseSection`
- :class:`~robot.parsing.model.blocks.KeywordSection`
- :class:`~robot.parsing.model.blocks.CommentSection`
- :class:`~robot.parsing.model.blocks.TestCase`
- :class:`~robot.parsing.model.blocks.Keyword`
- :class:`~robot.parsing.model.blocks.If`
- :class:`~robot.parsing.model.blocks.Try`
- :class:`~robot.parsing.model.blocks.For`
- :class:`~robot.parsing.model.blocks.While`

Statements:

- :class:`~robot.parsing.model.statements.SectionHeader`
- :class:`~robot.parsing.model.statements.LibraryImport`
- :class:`~robot.parsing.model.statements.ResourceImport`
- :class:`~robot.parsing.model.statements.VariablesImport`
- :class:`~robot.parsing.model.statements.Documentation`
- :class:`~robot.parsing.model.statements.Metadata`
- :class:`~robot.parsing.model.statements.ForceTags`
- :class:`~robot.parsing.model.statements.DefaultTags`
- :class:`~robot.parsing.model.statements.SuiteName`
- :class:`~robot.parsing.model.statements.SuiteSetup`
- :class:`~robot.parsing.model.statements.SuiteTeardown`
- :class:`~robot.parsing.model.statements.TestSetup`
- :class:`~robot.parsing.model.statements.TestTeardown`
- :class:`~robot.parsing.model.statements.TestTemplate`
- :class:`~robot.parsing.model.statements.TestTimeout`
- :class:`~robot.parsing.model.statements.Variable`
- :class:`~robot.parsing.model.statements.TestCaseName`
- :class:`~robot.parsing.model.statements.KeywordName`
- :class:`~robot.parsing.model.statements.Setup`
- :class:`~robot.parsing.model.statements.Teardown`
- :class:`~robot.parsing.model.statements.Tags`
- :class:`~robot.parsing.model.statements.Template`
- :class:`~robot.parsing.model.statements.Timeout`
- :class:`~robot.parsing.model.statements.Arguments`
- :class:`~robot.parsing.model.statements.Return` (deprecated, will mean ``ReturnStatement`` in RF 7.0)
- :class:`~robot.parsing.model.statements.ReturnSetting` (alias for ``Return``, new in RF 6.1)
- :class:`~robot.parsing.model.statements.KeywordCall`
- :class:`~robot.parsing.model.statements.TemplateArguments`
- :class:`~robot.parsing.model.statements.IfHeader`
- :class:`~robot.parsing.model.statements.InlineIfHeader`
- :class:`~robot.parsing.model.statements.ElseIfHeader`
- :class:`~robot.parsing.model.statements.ElseHeader`
- :class:`~robot.parsing.model.statements.TryHeader`
- :class:`~robot.parsing.model.statements.ExceptHeader`
- :class:`~robot.parsing.model.statements.FinallyHeader`
- :class:`~robot.parsing.model.statements.ForHeader`
- :class:`~robot.parsing.model.statements.WhileHeader`
- :class:`~robot.parsing.model.statements.End`
- :class:`~robot.parsing.model.statements.ReturnStatement`
- :class:`~robot.parsing.model.statements.Break`
- :class:`~robot.parsing.model.statements.Continue`
- :class:`~robot.parsing.model.statements.Comment`
- :class:`~robot.parsing.model.statements.Config` (new in RF 6.0)
- :class:`~robot.parsing.model.statements.Error`
- :class:`~robot.parsing.model.statements.EmptyLine`

Inspecting model
----------------

The easiest way to inspect what data a model contains is implementing
:class:`~robot.parsing.model.visitor.ModelVisitor` and creating
``visit_NodeName`` to visit nodes with name ``NodeName`` as needed.
The following example illustrates how to find what tests a certain test
case file contains::

    from robot.api.parsing import get_model, ModelVisitor


    class TestNamePrinter(ModelVisitor):

        def visit_File(self, node):
            print(f"File '{node.source}' has the following tests:")
            # Call `generic_visit` to visit also child nodes.
            self.generic_visit(node)

        def visit_TestCaseName(self, node):
            print(f"- {node.name} (on line {node.lineno})")


    model = get_model('example.robot')
    printer = TestNamePrinter()
    printer.visit(model)

When the above code is run using the earlier :file:`example.robot`, the
output is this::

    File 'example.robot' has the following tests:
    - Example (on line 2)
    - Second example (on line 5)

Handling errors in model
------------------------

All nodes in the model have ``errors`` attribute that contains possible errors
the node has. These errors include syntax errors such as empty FOR loops or IF
without a condition as well as errors caused by unrecognized data such as
non-existing section or setting names.

Unrecognized data is handled already during the tokenizing__ phase. In the model
such data is represented as :class:`~robot.parsing.model.statements.Error`
nodes and their ``errors`` attribute contain error information got from the
underlying ``ERROR`` tokens. Syntax errors do not create
:class:`~robot.parsing.model.statements.Error`
nodes, but instead the model has normal nodes such as
:class:`~robot.parsing.model.blocks.If`
with errors in their ``errors`` attribute.

A simple way to go through the model and see are there errors is using the
:class:`~robot.parsing.model.visitor.ModelVisitor`
discussed in the previous section::

    class ErrorReporter(ModelVisitor):

        # Implement `generic_visit` to visit all nodes.
        def generic_visit(self, node):
            if node.errors:
                print(f'Error on line {node.lineno}:')
                for error in node.errors:
                    print(f'- {error}')
            ModelVisitor.generic_visit(self, node)

__ `Parsing data to tokens`_

Modifying data
--------------

Existing data the model contains can be modified simply by modifying values of
the underlying tokens. If changes need to be saved, that is as easy as calling
the :meth:`~.model.blocks.File.save` method of the root model object. When
just modifying token values, it is possible to still use
:class:`~robot.parsing.model.visitor.ModelVisitor`
discussed in the above section. The next section discusses adding or removing
nodes and then
:class:`~robot.parsing.model.visitor.ModelTransformer`
should be used instead.

Modifications to tokens obviously require finding the tokens to be modified.
The first step is finding nodes containing the tokens by implementing
needed ``visit_NodeName`` methods. Then the exact token or tokens
can be found using nodes'
:meth:`~.model.statements.Statement.get_token` or
:meth:`~.model.statements.Statement.get_tokens` methods.
If only token values are needed,
:meth:`~.model.statements.Statement.get_value` or
:meth:`~.model.statements.Statement.get_values` can be used as a shortcut.
First finding nodes and then the right tokens is illustrated by
this keyword renaming example::

    from robot.api.parsing import get_model, ModelVisitor, Token


    class KeywordRenamer(ModelVisitor):

        def __init__(self, old_name, new_name):
            self.old_name = self.normalize(old_name)
            self.new_name = new_name

        def normalize(self, name):
            return name.lower().replace(' ', '').replace('_', '')

        def visit_KeywordName(self, node):
            '''Rename keyword definitions.'''
            if self.normalize(node.name) == self.old_name:
                token = node.get_token(Token.KEYWORD_NAME)
                token.value = self.new_name

        def visit_KeywordCall(self, node):
            '''Rename keyword usages.'''
            if self.normalize(node.keyword) == self.old_name:
                token = node.get_token(Token.KEYWORD)
                token.value = self.new_name


    model = get_model('example.robot')
    renamer = KeywordRenamer('Keyword', 'New Name')
    renamer.visit(model)
    model.save()

If you run the above example using the earlier :file:`example.robot`, you
can see that the ``Keyword`` keyword has been renamed to ``New Name``. Notice
that a real keyword renamer needed to take into account also keywords used
with setups, teardowns and templates.

When token values are changed, column offset of the other tokens on same
line are likely to be wrong. This does not affect saving the model or other
typical usages, but if it is a problem then the caller needs to updated
offsets separately.

Adding and removing nodes
-------------------------

Bigger changes to the model are somewhat more complicated than just modifying
existing token values. When doing this kind of changes,
:class:`~robot.parsing.model.visitor.ModelTransformer`
should be used instead of
:class:`~robot.parsing.model.visitor.ModelVisitor`
that was discussed in the previous sections.

Removing nodes is relative easy and is accomplished by returning ``None``
from ``visit_NodeName`` methods. Remember to return the original node,
or possibly a replacement node, from all of these methods when you do not
want a node to be removed.

Adding nodes requires constructing needed `Model objects`_ and adding them
to the model. The following example demonstrates both removing and adding nodes.
If you run it against the earlier :file:`example.robot`, you see that
the first test gets a new keyword, the second test is removed, and
settings section with documentation is added.

::

    from robot.api.parsing import (
        get_model, Documentation, EmptyLine, KeywordCall,
        ModelTransformer, SettingSection, SectionHeader, Token
    )


    class TestModifier(ModelTransformer):

        def visit_TestCase(self, node):
            # The matched `TestCase` node is a block with `header` and
            # `body` attributes. `header` is a statement with familiar
            # `get_token` and `get_value` methods for getting certain
            # tokens or their value.
            name = node.header.get_value(Token.TESTCASE_NAME)
            # Returning `None` drops the node altogether i.e. removes
            # this test.
            if name == 'Second example':
                return None
            # Construct new keyword call statement from tokens. See `visit_File`
            # below for an example creating statements using `from_params`.
            new_keyword = KeywordCall([
                Token(Token.SEPARATOR, '    '),
                Token(Token.KEYWORD, 'New Keyword'),
                Token(Token.SEPARATOR, '    '),
                Token(Token.ARGUMENT, 'xxx'),
                Token(Token.EOL)
            ])
            # Add the keyword call to test as the second item.
            node.body.insert(1, new_keyword)
            # No need to call `generic_visit` because we are not
            # modifying child nodes. The node itself must to be
            # returned to avoid dropping it.
            return node

        def visit_File(self, node):
            # Create settings section with documentation. Needed header and body
            # statements are created using `from_params` method. This is typically
            # more convenient than creating statements based on tokens like above.
            settings = SettingSection(
                header=SectionHeader.from_params(Token.SETTING_HEADER),
                body=[
                    Documentation.from_params('This is a really\npowerful API!'),
                    EmptyLine.from_params()
                ]
            )
            # Add settings to the beginning of the file.
            node.sections.insert(0, settings)
            # Call `generic_visit` to visit also child nodes.
            return self.generic_visit(node)


    model = get_model('example.robot')
    TestModifier().visit(model)
    model.save('modified.robot')

Executing model
---------------

It is possible to convert a parsed and possibly modified model into an
executable :class:`~robot.running.model.TestSuite` structure by using its
:func:`~robot.running.model.TestSuite.from_model` class method. In this case
the :func:`~.parser.parser.get_model` function should be given the ``curdir``
argument to get possible ``${CURDIR}`` variable resolved correctly.

::

     from robot.api import TestSuite
     from robot.api.parsing import get_model

     model = get_model('example.robot', curdir='/home/robot/example')
     # modify model as needed
     suite = TestSuite.from_model(model)
     suite.run()

For more details about executing the created
:class:`~robot.running.model.TestSuite` object, see the documentation
of its :meth:`~robot.running.model.TestSuite.run` method. Notice also
that if you do not need to modify the parsed model, it is easier to
get the executable suite by using the
:func:`~robot.running.model.TestSuite.from_file_system` class method.
"""

from robot.parsing import (
    get_tokens,
    get_resource_tokens,
    get_init_tokens,
    get_model,
    get_resource_model,
    get_init_model,
    Token
)
from robot.parsing.model.blocks import (
    File,
    SettingSection,
    VariableSection,
    TestCaseSection,
    KeywordSection,
    CommentSection,
    TestCase,
    Keyword,
    If,
    Try,
    For,
    While
)
from robot.parsing.model.statements import (
    SectionHeader,
    LibraryImport,
    ResourceImport,
    VariablesImport,
    Documentation,
    Metadata,
    SuiteName,
    SuiteSetup,
    SuiteTeardown,
    TestSetup,
    TestTeardown,
    TestTemplate,
    TestTimeout,
    ForceTags,
    DefaultTags,
    KeywordTags,
    Variable,
    TestCaseName,
    KeywordName,
    Setup,
    Teardown,
    Tags,
    Template,
    Timeout,
    Arguments,
    Return,
    ReturnSetting,
    KeywordCall,
    TemplateArguments,
    IfHeader,
    InlineIfHeader,
    ElseIfHeader,
    ElseHeader,
    TryHeader,
    ExceptHeader,
    FinallyHeader,
    ForHeader,
    WhileHeader,
    End,
    ReturnStatement,
    Continue,
    Break,
    Comment,
    Config,
    Error,
    EmptyLine
)
from robot.parsing.model.visitor import (
    ModelTransformer,
    ModelVisitor
)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import os.path

from robot.running import ArgInfo
from robot.errors import DataError

from .datatypes import EnumMember, TypedDictItem, TypeDoc
from .model import LibraryDoc, KeywordDoc


class JsonDocBuilder:

    def build(self, path):
        spec = self._parse_spec_json(path)
        return self.build_from_dict(spec)

    def build_from_dict(self, spec):
        libdoc = LibraryDoc(name=spec['name'],
                            doc=spec['doc'],
                            version=spec['version'],
                            type=spec['type'],
                            scope=spec['scope'],
                            doc_format=spec['docFormat'],
                            source=spec['source'],
                            lineno=int(spec.get('lineno', -1)))
        libdoc.inits = [self._create_keyword(kw) for kw in spec['inits']]
        libdoc.keywords = [self._create_keyword(kw) for kw in spec['keywords']]
        # RF >= 5 have 'typedocs', RF >= 4 have 'dataTypes', older/custom may have neither.
        if 'typedocs' in spec:
            libdoc.type_docs = self._parse_type_docs(spec['typedocs'])
        elif 'dataTypes' in spec:
            libdoc.type_docs = self._parse_data_types(spec['dataTypes'])
        return libdoc

    def _parse_spec_json(self, path):
        if not os.path.isfile(path):
            raise DataError(f"Spec file '{path}' does not exist.")
        with open(path) as json_source:
            libdoc_dict = json.load(json_source)
        return libdoc_dict

    def _create_keyword(self, data):
        kw = KeywordDoc(name=data.get('name'),
                        doc=data['doc'],
                        shortdoc=data['shortdoc'],
                        tags=data['tags'],
                        private=data.get('private', False),
                        deprecated=data.get('deprecated', False),
                        source=data['source'],
                        lineno=int(data.get('lineno', -1)))
        self._create_arguments(data['args'], kw)
        return kw

    def _create_arguments(self, arguments, kw: KeywordDoc):
        spec = kw.args
        setters = {
            ArgInfo.POSITIONAL_ONLY: spec.positional_only.append,
            ArgInfo.POSITIONAL_ONLY_MARKER: lambda value: None,
            ArgInfo.POSITIONAL_OR_NAMED: spec.positional_or_named.append,
            ArgInfo.VAR_POSITIONAL: lambda value: setattr(spec, 'var_positional', value),
            ArgInfo.NAMED_ONLY_MARKER: lambda value: None,
            ArgInfo.NAMED_ONLY: spec.named_only.append,
            ArgInfo.VAR_NAMED: lambda value: setattr(spec, 'var_named', value),
        }
        for arg in arguments:
            name = arg['name']
            setters[arg['kind']](name)
            default = arg.get('defaultValue')
            if default is not None:
                spec.defaults[name] = default
            if arg.get('type'):
                type_docs = {}
                type_info = self._parse_modern_type_info(arg['type'], type_docs)
            else:    # Compatibility with RF < 6.1.
                type_docs = arg.get('typedocs', {})
                type_info = tuple(arg['types'])
            if type_info:
                if not spec.types:
                    spec.types = {}
                spec.types[name] = type_info
            kw.type_docs[name] = type_docs

    def _parse_modern_type_info(self, data, type_docs):
        if data.get('typedoc'):
            type_docs[data['name']] = data['typedoc']
        return {'name': data['name'],
                'nested': [self._parse_modern_type_info(nested, type_docs)
                           for nested in data.get('nested', ())]}

    def _parse_type_docs(self, type_docs):
        for data in type_docs:
            doc = TypeDoc(data['type'], data['name'], data['doc'], data['accepts'],
                          data['usages'])
            if doc.type == TypeDoc.ENUM:
                doc.members = [EnumMember(d['name'], d['value'])
                               for d in data['members']]
            if doc.type == TypeDoc.TYPED_DICT:
                doc.items = [TypedDictItem(d['key'], d['type'], d['required'])
                             for d in data['items']]
            yield doc

    # Code below used for parsing legacy 'dataTypes'.

    def _parse_data_types(self, data_types):
        for obj in data_types['enums']:
            yield self._create_enum_doc(obj)
        for obj in data_types['typedDicts']:
            yield self._create_typed_dict_doc(obj)

    def _create_enum_doc(self, data):
        return TypeDoc(TypeDoc.ENUM, data['name'], data['doc'],
                       members=[EnumMember(member['name'], member['value'])
                                for member in data['members']])

    def _create_typed_dict_doc(self, data):
        return TypeDoc(TypeDoc.TYPED_DICT, data['name'], data['doc'],
                       items=[TypedDictItem(item['key'], item['type'], item['required'])
                              for item in data['items']])



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.utils import is_list_like, type_name

from .filesetter import VariableFileSetter
from .replacer import VariableReplacer
from .store import VariableStore
from .tablesetter import VariableTableSetter


class Variables:
    """Represents a set of variables.

    Contains methods for replacing variables from list, scalars, and strings.
    On top of ${scalar}, @{list} and &{dict} variables, these methods handle
    also %{environment} variables.
    """

    def __init__(self):
        self.store = VariableStore(self)
        self._replacer = VariableReplacer(self)

    def __setitem__(self, name, value):
        self.store.add(name, value)

    def __getitem__(self, name):
        return self.store.get(name)

    def __contains__(self, name):
        return name in self.store

    def resolve_delayed(self):
        self.store.resolve_delayed()

    def replace_list(self, items, replace_until=None, ignore_errors=False):
        if not is_list_like(items):
            raise ValueError("'replace_list' requires list-like input, "
                             "got %s." % type_name(items))
        return self._replacer.replace_list(items, replace_until, ignore_errors)

    def replace_scalar(self, item, ignore_errors=False):
        return self._replacer.replace_scalar(item, ignore_errors)

    def replace_string(self, item, custom_unescaper=None, ignore_errors=False):
        return self._replacer.replace_string(item, custom_unescaper, ignore_errors)

    def set_from_file(self, path_or_variables, args=None, overwrite=False):
        setter = VariableFileSetter(self.store)
        return setter.set(path_or_variables, args, overwrite)

    def set_from_variable_table(self, variables, overwrite=False):
        setter = VariableTableSetter(self.store)
        setter.set(variables, overwrite)

    def clear(self):
        self.store.clear()

    def copy(self):
        variables = Variables()
        variables.store.data = self.store.data.copy()
        return variables

    def update(self, variables):
        self.store.update(variables.store)

    def as_dict(self, decoration=True):
        return self.store.as_dict(decoration=decoration)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import datetime
import os
import time

from robot.utils import file_writer


class LibdocOutput:

    def __init__(self, output_path, format):
        self._output_path = output_path
        self._format = format.upper()
        self._output_file = None

    def __enter__(self):
        if self._format == 'HTML':
            self._output_file = file_writer(self._output_path,
                                            usage='Libdoc output')
            return self._output_file
        return self._output_path

    def __exit__(self, *exc_info):
        if self._output_file:
            self._output_file.close()
        if any(exc_info):
            try:
                os.remove(self._output_path)
            except OSError:
                pass


def get_generation_time():
    """Return a timestamp that honors `SOURCE_DATE_EPOCH`.

    This timestamp is to be used for embedding in output files, so
    that builds can be made reproducible.
    """
    ts = float(os.getenv('SOURCE_DATE_EPOCH', time.time()))
    dt = datetime.datetime.fromtimestamp(round(ts), datetime.timezone.utc)
    return dt.isoformat()



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.utils import file_writer


class LibdocJsonWriter:

    def write(self, libdoc, outfile):
        with file_writer(outfile) as writer:
            writer.write(libdoc.to_json(indent=2))



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ast import literal_eval
from collections import OrderedDict
from collections.abc import ByteString, Container, Mapping, Sequence, Set
from datetime import datetime, date, timedelta
from decimal import InvalidOperation, Decimal
from enum import Enum
from numbers import Integral, Real
from os import PathLike
from pathlib import Path, PurePath
from typing import Any, Tuple, TypeVar, Union

from robot.conf import Languages
from robot.libraries.DateTime import convert_date, convert_time
from robot.utils import (eq, get_error_message, has_args, is_string, is_union,
                         plural_or_not as s, safe_str, seq2str, type_name, type_repr,
                         typeddict_types)


NoneType = type(None)


class TypeConverter:
    type = None
    type_name = None
    abc = None
    aliases = ()
    value_types = (str,)
    doc = None
    _converters = OrderedDict()
    _type_aliases = {}

    def __init__(self, used_type, custom_converters=None, languages=None):
        self.used_type = used_type
        self.custom_converters = custom_converters
        self.languages = languages or Languages()

    @classmethod
    def register(cls, converter):
        cls._converters[converter.type] = converter
        for name in (converter.type_name,) + converter.aliases:
            if name is not None and not isinstance(name, property):
                cls._type_aliases[name.lower()] = converter.type
        return converter

    @classmethod
    def converter_for(cls, type_, custom_converters=None, languages=None):
        try:
            hash(type_)
        except TypeError:
            return None
        if isinstance(type_, str):
            try:
                type_ = cls._type_aliases[type_.lower()]
            except KeyError:
                return None
        used_type = type_
        if getattr(type_, '__origin__', None) and type_.__origin__ is not Union:
            type_ = type_.__origin__
        if custom_converters:
            info = custom_converters.get_converter_info(type_)
            if info:
                return CustomConverter(used_type, info)
        if type_ in cls._converters:
            return cls._converters[type_](used_type, custom_converters, languages)
        for converter in cls._converters.values():
            if converter.handles(type_):
                return converter(used_type, custom_converters, languages)
        return None

    @classmethod
    def handles(cls, type_):
        handled = (cls.type, cls.abc) if cls.abc else cls.type
        return isinstance(type_, type) and issubclass(type_, handled)

    def convert(self, name, value, explicit_type=True, strict=True, kind='Argument'):
        if self.no_conversion_needed(value):
            return value
        if not self._handles_value(value):
            return self._handle_error(name, value, kind, strict=strict)
        try:
            if not isinstance(value, str):
                return self._non_string_convert(value, explicit_type)
            return self._convert(value, explicit_type)
        except ValueError as error:
            return self._handle_error(name, value, kind, error, strict)

    def no_conversion_needed(self, value):
        used_type = getattr(self.used_type, '__origin__', self.used_type)
        try:
            return isinstance(value, used_type)
        except TypeError:
            # Used type wasn't a class. Compare to generic type instead.
            if self.type and self.type is not self.used_type:
                return isinstance(value, self.type)
            raise

    def _handles_value(self, value):
        return isinstance(value, self.value_types)

    def _non_string_convert(self, value, explicit_type=True):
        return self._convert(value, explicit_type)

    def _convert(self, value, explicit_type=True):
        raise NotImplementedError

    def _handle_error(self, name, value, kind, error=None, strict=True):
        if not strict:
            return value
        value_type = '' if isinstance(value, str) else f' ({type_name(value)})'
        value = safe_str(value)
        ending = f': {error}' if (error and error.args) else '.'
        if name is None:
            raise ValueError(
                f"{kind} '{value}'{value_type} "
                f"cannot be converted to {self.type_name}{ending}"
            )
        raise ValueError(
            f"{kind} '{name}' got value '{value}'{value_type} that "
            f"cannot be converted to {self.type_name}{ending}"
        )

    def _literal_eval(self, value, expected):
        if expected is set and value == 'set()':
            # `ast.literal_eval` has no way to define an empty set.
            return set()
        try:
            value = literal_eval(value)
        except (ValueError, SyntaxError):
            # Original errors aren't too informative in these cases.
            raise ValueError('Invalid expression.')
        except TypeError as err:
            raise ValueError(f'Evaluating expression failed: {err}')
        if not isinstance(value, expected):
            raise ValueError(f'Value is {type_name(value)}, not {expected.__name__}.')
        return value

    def _get_nested_types(self, type_hint, expected_count=None):
        types = getattr(type_hint, '__args__', ())
        # With generics from typing like Dict, __args__ is None with Python 3.6 and
        # contains TypeVars with 3.7-3.8. Newer versions don't have __args__ at all.
        # Subscripted usages like Dict[x, y].__args__ work fine with all.
        if not types or all(isinstance(a, TypeVar) for a in types):
            return ()
        if expected_count and len(types) != expected_count:
            raise TypeError(f'{type_hint.__name__}[] construct used as a type hint '
                            f'requires exactly {expected_count} nested '
                            f'type{s(expected_count)}, got {len(types)}.')
        return types

    def _remove_number_separators(self, value):
        if is_string(value):
            for sep in ' ', '_':
                if sep in value:
                    value = value.replace(sep, '')
        return value


@TypeConverter.register
class EnumConverter(TypeConverter):
    type = Enum

    @property
    def type_name(self):
        return self.used_type.__name__

    @property
    def value_types(self):
        return (str, int) if issubclass(self.used_type, int) else (str,)

    def _convert(self, value, explicit_type=True):
        enum = self.used_type
        if isinstance(value, int):
            return self._find_by_int_value(enum, value)
        try:
            return enum[value]
        except KeyError:
            return self._find_by_normalized_name_or_int_value(enum, value)

    def _find_by_normalized_name_or_int_value(self, enum, value):
        members = sorted(enum.__members__)
        matches = [m for m in members if eq(m, value, ignore='_')]
        if len(matches) == 1:
            return getattr(enum, matches[0])
        if len(matches) > 1:
            raise ValueError(f"{self.type_name} has multiple members matching "
                             f"'{value}'. Available: {seq2str(matches)}")
        try:
            if issubclass(self.used_type, int):
                return self._find_by_int_value(enum, value)
        except ValueError:
            members = [f'{m} ({getattr(enum, m)})' for m in members]
        raise ValueError(f"{self.type_name} does not have member '{value}'. "
                         f"Available: {seq2str(members)}")

    def _find_by_int_value(self, enum, value):
        value = int(value)
        for member in enum:
            if member.value == value:
                return member
        values = sorted(member.value for member in enum)
        raise ValueError(f"{self.type_name} does not have value '{value}'. "
                         f"Available: {seq2str(values)}")


@TypeConverter.register
class AnyConverter(TypeConverter):
    type = Any
    type_name = 'Any'
    aliases = ('any',)
    value_types = (Any,)

    @classmethod
    def handles(cls, type_):
        return type_ is Any

    def no_conversion_needed(self, value):
        return True

    def _convert(self, value, explicit_type=True):
        return value

    def _handles_value(self, value):
        return True


@TypeConverter.register
class StringConverter(TypeConverter):
    type = str
    type_name = 'string'
    aliases = ('string', 'str', 'unicode')
    value_types = (Any,)

    def _handles_value(self, value):
        return True

    def _convert(self, value, explicit_type=True):
        if not explicit_type:
            return value
        try:
            return str(value)
        except Exception:
            raise ValueError(get_error_message())


@TypeConverter.register
class BooleanConverter(TypeConverter):
    type = bool
    type_name = 'boolean'
    aliases = ('bool',)
    value_types = (str, int, float, NoneType)

    def _non_string_convert(self, value, explicit_type=True):
        return value

    def _convert(self, value, explicit_type=True):
        normalized = value.title()
        if normalized == 'None':
            return None
        if normalized in self.languages.true_strings:
            return True
        if normalized in self.languages.false_strings:
            return False
        return value


@TypeConverter.register
class IntegerConverter(TypeConverter):
    type = int
    abc = Integral
    type_name = 'integer'
    aliases = ('int', 'long')
    value_types = (str, float)

    def _non_string_convert(self, value, explicit_type=True):
        if value.is_integer():
            return int(value)
        raise ValueError('Conversion would lose precision.')

    def _convert(self, value, explicit_type=True):
        value = self._remove_number_separators(value)
        value, base = self._get_base(value)
        try:
            return int(value, base)
        except ValueError:
            if base == 10 and not explicit_type:
                try:
                    return float(value)
                except ValueError:
                    pass
        raise ValueError

    def _get_base(self, value):
        value = value.lower()
        for prefix, base in [('0x', 16), ('0o', 8), ('0b', 2)]:
            if prefix in value:
                parts = value.split(prefix)
                if len(parts) == 2 and parts[0] in ('', '-', '+'):
                    return ''.join(parts), base
        return value, 10


@TypeConverter.register
class FloatConverter(TypeConverter):
    type = float
    abc = Real
    type_name = 'float'
    aliases = ('double',)
    value_types = (str, Real)

    def _convert(self, value, explicit_type=True):
        try:
            return float(self._remove_number_separators(value))
        except ValueError:
            raise ValueError


@TypeConverter.register
class DecimalConverter(TypeConverter):
    type = Decimal
    type_name = 'decimal'
    value_types = (str, int, float)

    def _convert(self, value, explicit_type=True):
        try:
            return Decimal(self._remove_number_separators(value))
        except InvalidOperation:
            # With Python 3 error messages by decimal module are not very
            # useful and cannot be included in our error messages:
            # https://bugs.python.org/issue26208
            raise ValueError


@TypeConverter.register
class BytesConverter(TypeConverter):
    type = bytes
    abc = ByteString
    type_name = 'bytes'
    value_types = (str, bytearray)

    def _non_string_convert(self, value, explicit_type=True):
        return bytes(value)

    def _convert(self, value, explicit_type=True):
        try:
            return value.encode('latin-1')
        except UnicodeEncodeError as err:
            invalid = value[err.start:err.start+1]
            raise ValueError(f"Character '{invalid}' cannot be mapped to a byte.")


@TypeConverter.register
class ByteArrayConverter(TypeConverter):
    type = bytearray
    type_name = 'bytearray'
    value_types = (str, bytes)

    def _non_string_convert(self, value, explicit_type=True):
        return bytearray(value)

    def _convert(self, value, explicit_type=True):
        try:
            return bytearray(value, 'latin-1')
        except UnicodeEncodeError as err:
            invalid = value[err.start:err.start+1]
            raise ValueError(f"Character '{invalid}' cannot be mapped to a byte.")


@TypeConverter.register
class DateTimeConverter(TypeConverter):
    type = datetime
    type_name = 'datetime'
    value_types = (str, int, float)

    def _convert(self, value, explicit_type=True):
        return convert_date(value, result_format='datetime')


@TypeConverter.register
class DateConverter(TypeConverter):
    type = date
    type_name = 'date'

    def _convert(self, value, explicit_type=True):
        dt = convert_date(value, result_format='datetime')
        if dt.hour or dt.minute or dt.second or dt.microsecond:
            raise ValueError("Value is datetime, not date.")
        return dt.date()


@TypeConverter.register
class TimeDeltaConverter(TypeConverter):
    type = timedelta
    type_name = 'timedelta'
    value_types = (str, int, float)

    def _convert(self, value, explicit_type=True):
        return convert_time(value, result_format='timedelta')


@TypeConverter.register
class PathConverter(TypeConverter):
    type = Path
    abc = PathLike
    type_name = 'Path'
    value_types = (str, PurePath)

    def _convert(self, value, explicit_type=True):
        return Path(value)


@TypeConverter.register
class NoneConverter(TypeConverter):
    type = NoneType
    type_name = 'None'

    @classmethod
    def handles(cls, type_):
        return type_ in (NoneType, None)

    def _convert(self, value, explicit_type=True):
        if value.upper() == 'NONE':
            return None
        raise ValueError


@TypeConverter.register
class ListConverter(TypeConverter):
    type = list
    type_name = 'list'
    abc = Sequence
    value_types = (str, Sequence)

    def __init__(self, used_type, custom_converters=None, languages=None):
        super().__init__(used_type, custom_converters, languages)
        types = self._get_nested_types(used_type, expected_count=1)
        if not types:
            self.converter = None
        else:
            self.type_name = type_repr(used_type)
            self.converter = self.converter_for(types[0], custom_converters, languages)

    @classmethod
    def handles(cls, type_):
        # `type_ is not Tuple` is needed with Python 3.6.
        return super().handles(type_) and type_ is not Tuple

    def no_conversion_needed(self, value):
        if isinstance(value, str) or not super().no_conversion_needed(value):
            return False
        if not self.converter:
            return True
        return all(self.converter.no_conversion_needed(v) for v in value)

    def _non_string_convert(self, value, explicit_type=True):
        return self._convert_items(list(value), explicit_type)

    def _convert(self, value, explicit_type=True):
        return self._convert_items(self._literal_eval(value, list), explicit_type)

    def _convert_items(self, value, explicit_type):
        if not self.converter:
            return value
        return [self.converter.convert(i, v, explicit_type, kind='Item')
                for i, v in enumerate(value)]


@TypeConverter.register
class TupleConverter(TypeConverter):
    type = tuple
    type_name = 'tuple'
    value_types = (str, Sequence)

    def __init__(self, used_type, custom_converters=None, languages=None):
        super().__init__(used_type, custom_converters, languages)
        self.converters = ()
        self.homogenous = False
        types = self._get_nested_types(used_type)
        if not types:
            return
        if types[-1] is Ellipsis:
            types = types[:-1]
            if len(types) != 1:
                raise TypeError(f'Homogenous tuple used as a type hint requires '
                                f'exactly one nested type, got {len(types)}.')
            self.homogenous = True
        self.type_name = type_repr(used_type)
        self.converters = tuple(self.converter_for(t, custom_converters, languages)
                                or NullConverter() for t in types)

    def no_conversion_needed(self, value):
        if isinstance(value, str) or not super().no_conversion_needed(value):
            return False
        if not self.converters:
            return True
        if self.homogenous:
            return all(self.converters[0].no_conversion_needed(v) for v in value)
        if len(value) != len(self.converters):
            return False
        return all(c.no_conversion_needed(v) for c, v in zip(self.converters, value))

    def _non_string_convert(self, value, explicit_type=True):
        return self._convert_items(tuple(value), explicit_type)

    def _convert(self, value, explicit_type=True):
        return self._convert_items(self._literal_eval(value, tuple), explicit_type)

    def _convert_items(self, value, explicit_type):
        if not self.converters:
            return value
        if self.homogenous:
            conv = self.converters[0]
            return tuple(conv.convert(str(i), v, explicit_type, kind='Item')
                         for i, v in enumerate(value))
        if len(self.converters) != len(value):
            raise ValueError(f'Expected {len(self.converters)} '
                             f'item{s(self.converters)}, got {len(value)}.')
        return tuple(conv.convert(i, v, explicit_type, kind='Item')
                     for i, (conv, v) in enumerate(zip(self.converters, value)))


@TypeConverter.register
class TypedDictConverter(TypeConverter):
    type = 'TypedDict'
    value_types = (str, Mapping)

    def __init__(self, used_type, custom_converters, languages=None):
        super().__init__(used_type, custom_converters, languages)
        self.converters = {n: self.converter_for(t, custom_converters, languages)
                           for n, t in used_type.__annotations__.items()}
        self.type_name = used_type.__name__
        # __required_keys__ is new in Python 3.9.
        self.required_keys = getattr(used_type, '__required_keys__', frozenset())

    @classmethod
    def handles(cls, type_):
        return isinstance(type_, typeddict_types)

    def no_conversion_needed(self, value):
        return False

    def _non_string_convert(self, value, explicit_type=True):
        return self._convert_items(value)

    def _convert(self, value, explicit_type=True):
        return self._convert_items(self._literal_eval(value, dict))

    def _convert_items(self, value):
        not_allowed = []
        for key in value:
            try:
                converter = self.converters[key]
            except KeyError:
                not_allowed.append(key)
            else:
                if converter:
                    value[key] = converter.convert(key, value[key], kind='Item')
        if not_allowed:
            error = f'Item{s(not_allowed)} {seq2str(sorted(not_allowed))} not allowed.'
            available = [key for key in self.converters if key not in value]
            if available:
                error += f' Available item{s(available)}: {seq2str(sorted(available))}'
            raise ValueError(error)
        missing = [key for key in self.required_keys if key not in value]
        if missing:
            raise ValueError(f"Required item{s(missing)} "
                             f"{seq2str(sorted(missing))} missing.")
        return value


@TypeConverter.register
class DictionaryConverter(TypeConverter):
    type = dict
    abc = Mapping
    type_name = 'dictionary'
    aliases = ('dict', 'map')
    value_types = (str, Mapping)

    def __init__(self, used_type, custom_converters=None, languages=None):
        super().__init__(used_type, custom_converters, languages)
        types = self._get_nested_types(used_type, expected_count=2)
        if not types:
            self.converters = ()
        else:
            self.type_name = type_repr(used_type)
            self.converters = tuple(self.converter_for(t, custom_converters, languages)
                                    or NullConverter() for t in types)

    def no_conversion_needed(self, value):
        if isinstance(value, str) or not super().no_conversion_needed(value):
            return False
        if not self.converters:
            return True
        no_key_conversion_needed = self.converters[0].no_conversion_needed
        no_value_conversion_needed = self.converters[1].no_conversion_needed
        return all(no_key_conversion_needed(k) and no_value_conversion_needed(v)
                   for k, v in value.items())

    def _non_string_convert(self, value, explicit_type=True):
        if self._used_type_is_dict() and not isinstance(value, dict):
            value = dict(value)
        return self._convert_items(value, explicit_type)

    def _used_type_is_dict(self):
        used_type = getattr(self.used_type, '__origin__', self.used_type)
        return issubclass(used_type, dict)

    def _convert(self, value, explicit_type=True):
        return self._convert_items(self._literal_eval(value, dict), explicit_type)

    def _convert_items(self, value, explicit_type):
        if not self.converters:
            return value
        convert_key = self._get_converter(self.converters[0], explicit_type, 'Key')
        convert_value = self._get_converter(self.converters[1], explicit_type, 'Item')
        return {convert_key(None, k): convert_value(k, v) for k, v in value.items()}

    def _get_converter(self, converter, explicit_type, kind):
        return lambda name, value: converter.convert(name, value, explicit_type,
                                                     kind=kind)


@TypeConverter.register
class SetConverter(TypeConverter):
    type = set
    abc = Set
    type_name = 'set'
    value_types = (str, Container)

    def __init__(self, used_type, custom_converters=None, languages=None):
        super().__init__(used_type, custom_converters, languages)
        types = self._get_nested_types(used_type, expected_count=1)
        if not types:
            self.converter = None
        else:
            self.type_name = type_repr(used_type)
            self.converter = self.converter_for(types[0], custom_converters, languages)

    def no_conversion_needed(self, value):
        if isinstance(value, str) or not super().no_conversion_needed(value):
            return False
        if not self.converter:
            return True
        return all(self.converter.no_conversion_needed(v) for v in value)

    def _non_string_convert(self, value, explicit_type=True):
        return self._convert_items(set(value), explicit_type)

    def _convert(self, value, explicit_type=True):
        return self._convert_items(self._literal_eval(value, set), explicit_type)

    def _convert_items(self, value, explicit_type):
        if not self.converter:
            return value
        return {self.converter.convert(None, v, explicit_type, kind='Item')
                for v in value}


@TypeConverter.register
class FrozenSetConverter(SetConverter):
    type = frozenset
    type_name = 'frozenset'

    def _non_string_convert(self, value, explicit_type=True):
        return frozenset(super()._non_string_convert(value, explicit_type))

    def _convert(self, value, explicit_type=True):
        # There are issues w/ literal_eval. See self._literal_eval for details.
        if value == 'frozenset()':
            return frozenset()
        return frozenset(super()._convert(value, explicit_type))


@TypeConverter.register
class CombinedConverter(TypeConverter):
    type = Union

    def __init__(self, union, custom_converters, languages=None):
        super().__init__(self._get_types(union))
        self.converters = tuple(self.converter_for(t, custom_converters, languages)
                                for t in self.used_type)

    def _get_types(self, union):
        if not union:
            return ()
        if isinstance(union, tuple):
            return union
        if has_args(union):
            return union.__args__
        return ()

    @property
    def type_name(self):
        if not self.used_type:
            return 'union'
        return ' or '.join(type_name(t) for t in self.used_type)

    @classmethod
    def handles(cls, type_):
        return is_union(type_, allow_tuple=True)

    def _handles_value(self, value):
        return True

    def no_conversion_needed(self, value):
        for converter, type_ in zip(self.converters, self.used_type):
            if converter:
                if converter.no_conversion_needed(value):
                    return True
            else:
                try:
                    if isinstance(value, type_):
                        return True
                except TypeError:
                    pass
        return False

    def _convert(self, value, explicit_type=True):
        if not self.used_type:
            raise ValueError('Cannot have union without types.')
        unrecognized_types = False
        for converter in self.converters:
            if converter:
                try:
                    return converter.convert('', value, explicit_type)
                except ValueError:
                    pass
            else:
                unrecognized_types = True
        if unrecognized_types:
            return value
        raise ValueError


class CustomConverter(TypeConverter):

    def __init__(self, used_type, converter_info, languages=None):
        super().__init__(used_type, languages=languages)
        self.converter_info = converter_info

    @property
    def type_name(self):
        return self.converter_info.name

    @property
    def doc(self):
        return self.converter_info.doc

    @property
    def value_types(self):
        return self.converter_info.value_types

    def _handles_value(self, value):
        return not self.value_types or isinstance(value, self.value_types)

    def _convert(self, value, explicit_type=True):
        try:
            return self.converter_info.convert(value)
        except ValueError:
            raise
        except Exception:
            raise ValueError(get_error_message())


class NullConverter:

    def convert(self, name, value, explicit_type=True, strict=True, kind='Argument'):
        return value

    def no_conversion_needed(self, value):
        return True



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

from robot.utils import format_time
from .loggerhelper import Message, write_to_console


class StdoutLogSplitter:
    """Splits messages logged through stdout (or stderr) into Message objects"""

    _split_from_levels = re.compile(r'^(?:\*'
                                    r'(TRACE|DEBUG|INFO|CONSOLE|HTML|WARN|ERROR)'
                                    r'(:\d+(?:\.\d+)?)?'  # Optional timestamp
                                    r'\*)', re.MULTILINE)

    def __init__(self, output):
        self._messages = list(self._get_messages(output.strip()))

    def _get_messages(self, output):
        for level, timestamp, msg in self._split_output(output):
            if level == 'CONSOLE':
                write_to_console(msg.lstrip())
                level = 'INFO'
            if timestamp:
                timestamp = self._format_timestamp(timestamp[1:])
            yield Message(msg.strip(), level, timestamp=timestamp)

    def _split_output(self, output):
        tokens = self._split_from_levels.split(output)
        tokens = self._add_initial_level_and_time_if_needed(tokens)
        for i in range(0, len(tokens), 3):
            yield tokens[i:i+3]

    def _add_initial_level_and_time_if_needed(self, tokens):
        if self._output_started_with_level(tokens):
            return tokens[1:]
        return ['INFO', None] + tokens

    def _output_started_with_level(self, tokens):
        return tokens[0] == ''

    def _format_timestamp(self, millis):
        return format_time(float(millis)/1000, millissep='.')

    def __iter__(self):
        return iter(self._messages)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import json
from pathlib import Path
from typing import Any, Dict, overload, TextIO, Type, TypeVar

from robot.errors import DataError
from robot.utils import get_error_message, SetterAwareType, type_name


T = TypeVar('T', bound='ModelObject')
DataDict = Dict[str, Any]


class ModelObject(metaclass=SetterAwareType):
    repr_args = ()
    __slots__ = []

    @classmethod
    def from_dict(cls: Type[T], data: DataDict) -> T:
        """Create this object based on data in a dictionary.

        Data can be got from the :meth:`to_dict` method or created externally.
        """
        try:
            return cls().config(**data)
        except (AttributeError, TypeError) as err:
            raise DataError(f"Creating '{full_name(cls)}' object from dictionary "
                            f"failed: {err}")

    @classmethod
    def from_json(cls: Type[T], source: 'str|bytes|TextIO|Path') -> T:
        """Create this object based on JSON data.

        The data is given as the ``source`` parameter. It can be:

        - a string (or bytes) containing the data directly,
        - an open file object where to read the data, or
        - a path (``pathlib.Path`` or string) to a UTF-8 encoded file to read.

        The JSON data is first converted to a Python dictionary and the object
        created using the :meth:`from_dict` method.

        Notice that the ``source`` is considered to be JSON data if it is
        a string and contains ``{``. If you need to use ``{`` in a file system
        path, pass it in as a ``pathlib.Path`` instance.
        """
        try:
            data = JsonLoader().load(source)
        except (TypeError, ValueError) as err:
            raise DataError(f'Loading JSON data failed: {err}')
        return cls.from_dict(data)

    def to_dict(self) -> DataDict:
        """Serialize this object into a dictionary.

        The object can be later restored by using the :meth:`from_dict` method.
        """
        raise NotImplementedError

    @overload
    def to_json(self, file: None = None, *, ensure_ascii: bool = False,
                indent: int = 0, separators: 'tuple[str, str]' = (',', ':')) -> str:
        ...

    @overload
    def to_json(self, file: 'TextIO|Path|str', *, ensure_ascii: bool = False,
                indent: int = 0, separators: 'tuple[str, str]' = (',', ':')) -> None:
        ...

    def to_json(self, file: 'None|TextIO|Path|str' = None, *,
                ensure_ascii: bool = False, indent: int = 0,
                separators: 'tuple[str, str]' = (',', ':')) -> 'None|str':
        """Serialize this object into JSON.

        The object is first converted to a Python dictionary using the
        :meth:`to_dict` method and then the dictionary is converted to JSON.

        The ``file`` parameter controls what to do with the resulting JSON data.
        It can be:

        - ``None`` (default) to return the data as a string,
        - an open file object where to write the data, or
        - a path (``pathlib.Path`` or string) to a file where to write
          the data using UTF-8 encoding.

        JSON formatting can be configured using optional parameters that
        are passed directly to the underlying json__ module. Notice that
        the defaults differ from what ``json`` uses.

        __ https://docs.python.org/3/library/json.html
        """
        return JsonDumper(ensure_ascii=ensure_ascii, indent=indent,
                          separators=separators).dump(self.to_dict(), file)

    def config(self: T, **attributes) -> T:
        """Configure model object with given attributes.

        ``obj.config(name='Example', doc='Something')`` is equivalent to setting
        ``obj.name = 'Example'`` and ``obj.doc = 'Something'``.

        New in Robot Framework 4.0.
        """
        for name, value in attributes.items():
            try:
                orig = getattr(self, name)
            except AttributeError:
                raise AttributeError(f"'{full_name(self)}' object does not have "
                                     f"attribute '{name}'")
            # Preserve tuples. Main motivation is converting lists with `from_json`.
            if isinstance(orig, tuple) and not isinstance(value, tuple):
                try:
                    value = tuple(value)
                except TypeError:
                    raise TypeError(f"'{full_name(self)}' object attribute '{name}' "
                                    f"is 'tuple', got '{type_name(value)}'.")
            try:
                setattr(self, name, value)
            except AttributeError as err:
                # Ignore error setting attribute if the object already has it.
                # Avoids problems with `from_dict` with body items having
                # un-settable `type` attribute that is needed in dict data.
                if value != orig:
                    raise AttributeError(f"Setting attribute '{name}' failed: {err}")
        return self

    def copy(self: T, **attributes) -> T:
        """Return a shallow copy of this object.

        :param attributes: Attributes to be set to the returned copy.
            For example, ``obj.copy(name='New name')``.

        See also :meth:`deepcopy`. The difference between ``copy`` and
        ``deepcopy`` is the same as with the methods having same names in
        the copy__ module.

        __ https://docs.python.org/3/library/copy.html
        """
        return copy.copy(self).config(**attributes)

    def deepcopy(self: T, **attributes) -> T:
        """Return a deep copy of this object.

        :param attributes: Attributes to be set to the returned copy.
            For example, ``obj.deepcopy(name='New name')``.

        See also :meth:`copy`. The difference between ``deepcopy`` and
        ``copy`` is the same as with the methods having same names in
        the copy__ module.

        __ https://docs.python.org/3/library/copy.html
        """
        return copy.deepcopy(self).config(**attributes)

    def __repr__(self) -> str:
        arguments = [(name, getattr(self, name)) for name in self.repr_args]
        args_repr = ', '.join(f'{name}={value!r}' for name, value in arguments
                              if self._include_in_repr(name, value))
        return f"{full_name(self)}({args_repr})"

    def _include_in_repr(self, name: str, value: Any) -> bool:
        return True


def full_name(obj_or_cls):
    cls = type(obj_or_cls) if not isinstance(obj_or_cls, type) else obj_or_cls
    parts = cls.__module__.split('.') + [cls.__name__]
    if len(parts) > 1 and parts[0] == 'robot':
        parts[2:-1] = []
    return '.'.join(parts)


class JsonLoader:

    def load(self, source: 'str|bytes|TextIO|Path') -> DataDict:
        try:
            data = self._load(source)
        except (json.JSONDecodeError, TypeError):
            raise ValueError(f'Invalid JSON data: {get_error_message()}')
        if not isinstance(data, dict):
            raise TypeError(f"Expected dictionary, got {type_name(data)}.")
        return data

    def _load(self, source):
        if self._is_path(source):
            with open(source, encoding='UTF-8') as file:
                return json.load(file)
        if hasattr(source, 'read'):
            return json.load(source)
        return json.loads(source)

    def _is_path(self, source):
        if isinstance(source, Path):
            return True
        if not isinstance(source, str) or '{' in source:
            return False
        try:
            return Path(source).is_file()
        except OSError:    # Can happen on Windows w/ Python < 3.10.
            return False


class JsonDumper:

    def __init__(self, **config):
        self.config = config

    @overload
    def dump(self, data: DataDict, output: None = None) -> str:
        ...

    @overload
    def dump(self, data: DataDict, output: 'TextIO|Path|str') -> None:
        ...

    def dump(self, data: DataDict, output: 'None|TextIO|Path|str' = None) -> 'None|str':
        if not output:
            return json.dumps(data, **self.config)
        elif isinstance(output, (str, Path)):
            with open(output, 'w', encoding='UTF-8') as file:
                json.dump(data, file, **self.config)
        elif hasattr(output, 'write'):
            json.dump(data, output, **self.config)
        else:
            raise TypeError(f"Output should be None, path or open file, "
                            f"got {type_name(output)}.")



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from functools import partial

from robot.errors import VariableError
from robot.utils import is_dict_like, is_list_like, normalize, RecommendationFinder


def variable_not_found(name, candidates, message=None, deco_braces=True):
    """Raise DataError for missing variable name.

    Return recommendations for similar variable names if any are found.
    """
    candidates = _decorate_candidates(name[0], candidates, deco_braces)
    normalizer = partial(normalize, ignore='$@&%{}_')
    message = RecommendationFinder(normalizer).find_and_format(
        name, candidates,
        message=message or "Variable '%s' not found." % name
    )
    raise VariableError(message)


def _decorate_candidates(identifier, candidates, deco_braces=True):
    template = '%s{%s}' if deco_braces else '%s%s'
    is_included = {'$': lambda value: True,
                   '@': is_list_like,
                   '&': is_dict_like,
                   '%': lambda value: True}[identifier]
    return [template % (identifier, name)
            for name in candidates if is_included(candidates[name])]



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.errors import DataError
from robot.utils import get_error_message, is_bytes, is_list_like, is_string, type_name

from .arguments import PythonArgumentParser


def no_dynamic_method(*args):
    return None


class _DynamicMethod:
    _underscore_name = NotImplemented

    def __init__(self, lib):
        self.method = self._get_method(lib)

    def _get_method(self, lib):
        for name in self._underscore_name, self._camelCaseName:
            method = getattr(lib, name, None)
            if callable(method):
                return method
        return no_dynamic_method

    @property
    def _camelCaseName(self):
        tokens = self._underscore_name.split('_')
        return ''.join([tokens[0]] + [t.capitalize() for t in tokens[1:]])

    @property
    def name(self):
        return self.method.__name__

    def __call__(self, *args):
        try:
            return self._handle_return_value(self.method(*args))
        except:
            raise DataError("Calling dynamic method '%s' failed: %s"
                            % (self.name, get_error_message()))

    def _handle_return_value(self, value):
        raise NotImplementedError

    def _to_string(self, value, allow_tuple=False, allow_none=False):
        if is_string(value):
            return value
        if is_bytes(value):
            return value.decode('UTF-8')
        if allow_tuple and is_list_like(value) and len(value) > 0:
            return tuple(value)
        if allow_none and value is None:
            return value
        or_tuple = ' or a non-empty tuple' if allow_tuple else ''
        raise DataError('Return value must be a string%s, got %s.'
                        % (or_tuple, type_name(value)))

    def _to_list(self, value):
        if value is None:
            return ()
        if not is_list_like(value):
            raise DataError
        return value

    def _to_list_of_strings(self, value, allow_tuples=False):
        try:
            return [self._to_string(item, allow_tuples)
                    for item in self._to_list(value)]
        except DataError:
            raise DataError('Return value must be a list of strings%s.'
                            % (' or non-empty tuples' if allow_tuples else ''))

    def __bool__(self):
        return self.method is not no_dynamic_method


class GetKeywordNames(_DynamicMethod):
    _underscore_name = 'get_keyword_names'

    def _handle_return_value(self, value):
        names = self._to_list_of_strings(value)
        return list(self._remove_duplicates(names))

    def _remove_duplicates(self, names):
        seen = set()
        for name in names:
            if name not in seen:
                seen.add(name)
                yield name


class RunKeyword(_DynamicMethod):
    _underscore_name = 'run_keyword'

    @property
    def supports_kwargs(self):
        spec = PythonArgumentParser().parse(self.method)
        return len(spec.positional) == 3


class GetKeywordDocumentation(_DynamicMethod):
    _underscore_name = 'get_keyword_documentation'

    def _handle_return_value(self, value):
        return self._to_string(value or '')


class GetKeywordArguments(_DynamicMethod):
    _underscore_name = 'get_keyword_arguments'

    def __init__(self, lib):
        _DynamicMethod.__init__(self, lib)
        self._supports_kwargs = RunKeyword(lib).supports_kwargs

    def _handle_return_value(self, value):
        if value is None:
            if self._supports_kwargs:
                return ['*varargs', '**kwargs']
            return ['*varargs']
        return self._to_list_of_strings(value, allow_tuples=True)


class GetKeywordTypes(_DynamicMethod):
    _underscore_name = 'get_keyword_types'

    def _handle_return_value(self, value):
        return value if self else {}


class GetKeywordTags(_DynamicMethod):
    _underscore_name = 'get_keyword_tags'

    def _handle_return_value(self, value):
        return self._to_list_of_strings(value)


class GetKeywordSource(_DynamicMethod):
    _underscore_name = 'get_keyword_source'

    def _handle_return_value(self, value):
        return self._to_string(value, allow_none=True)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.conf import Languages, LanguageLike, LanguagesLike
from robot.utils import normalize_whitespace

from .settings import (InitFileSettings, FileSettings, Settings, SuiteFileSettings,
                       ResourceFileSettings, TestCaseSettings, KeywordSettings)
from .tokens import StatementTokens, Token


class LexingContext:

    def __init__(self, settings: Settings, languages: Languages):
        self.settings = settings
        self.languages = languages

    def lex_setting(self, statement: StatementTokens):
        self.settings.lex(statement)


class FileContext(LexingContext):
    settings: FileSettings

    def __init__(self, lang: LanguagesLike = None):
        languages = lang if isinstance(lang, Languages) else Languages(lang)
        settings_class: 'type[FileSettings]' = type(self).__annotations__['settings']
        settings = settings_class(languages)
        super().__init__(settings, languages)

    def add_language(self, lang: LanguageLike):
        self.languages.add_language(lang)

    def keyword_context(self) -> 'KeywordContext':
        return KeywordContext(KeywordSettings(self.settings))

    def setting_section(self, statement: StatementTokens) -> bool:
        return self._handles_section(statement, 'Settings')

    def variable_section(self, statement: StatementTokens) -> bool:
        return self._handles_section(statement, 'Variables')

    def test_case_section(self, statement: StatementTokens) -> bool:
        return False

    def task_section(self, statement: StatementTokens) -> bool:
        return False

    def keyword_section(self, statement: StatementTokens) -> bool:
        return self._handles_section(statement, 'Keywords')

    def comment_section(self, statement: StatementTokens) -> bool:
        return self._handles_section(statement, 'Comments')

    def lex_invalid_section(self, statement: StatementTokens):
        header = statement[0]
        header.type = Token.INVALID_HEADER
        header.error = self._get_invalid_section_error(header.value)
        for token in statement[1:]:
            token.type = Token.COMMENT

    def _get_invalid_section_error(self, header: str) -> str:
        raise NotImplementedError

    def _handles_section(self, statement: StatementTokens, header: str) -> bool:
        marker = statement[0].value
        return bool(marker and marker[0] == '*' and
                    self.languages.headers.get(self._normalize(marker)) == header)

    def _normalize(self, marker: str) -> str:
        return normalize_whitespace(marker).strip('* ').title()


class SuiteFileContext(FileContext):
    settings: SuiteFileSettings

    def test_case_context(self) -> 'TestCaseContext':
        return TestCaseContext(TestCaseSettings(self.settings))

    def test_case_section(self, statement: StatementTokens) -> bool:
        return self._handles_section(statement, 'Test Cases')

    def task_section(self, statement: StatementTokens) -> bool:
        return self._handles_section(statement, 'Tasks')

    def _get_invalid_section_error(self, header: str) -> str:
        return (f"Unrecognized section header '{header}'. Valid sections: "
                f"'Settings', 'Variables', 'Test Cases', 'Tasks', 'Keywords' "
                f"and 'Comments'.")


class ResourceFileContext(FileContext):
    settings: ResourceFileSettings

    def _get_invalid_section_error(self, header: str) -> str:
        name = self._normalize(header)
        if self.languages.headers.get(name) in ('Test Cases', 'Tasks'):
            return f"Resource file with '{name}' section is invalid."
        return (f"Unrecognized section header '{header}'. Valid sections: "
                f"'Settings', 'Variables', 'Keywords' and 'Comments'.")


class InitFileContext(FileContext):
    settings: InitFileSettings

    def _get_invalid_section_error(self, header: str) -> str:
        name = self._normalize(header)
        if self.languages.headers.get(name) in ('Test Cases', 'Tasks'):
            return f"'{name}' section is not allowed in suite initialization file."
        return (f"Unrecognized section header '{header}'. Valid sections: "
                f"'Settings', 'Variables', 'Keywords' and 'Comments'.")


class TestCaseContext(LexingContext):
    settings: TestCaseSettings

    def __init__(self, settings: TestCaseSettings):
        super().__init__(settings, settings.languages)

    @property
    def template_set(self) -> bool:
        return self.settings.template_set


class KeywordContext(LexingContext):
    settings: KeywordSettings

    def __init__(self, settings: KeywordSettings):
        super().__init__(settings, settings.languages)

    @property
    def template_set(self) -> bool:
        return False
